class AppTestMin:

    def test_min_notseq(self):
        raises(TypeError, min, 1)

    def test_min_usual(self):
        assert min(1, 2, 3) == 1

    def test_min_floats(self):
        assert min(0.1, 2.7, 14.7) == 0.1

    def test_min_chars(self):
        assert min('a', 'b', 'c') == 'a'

    def test_min_strings(self):
        assert min('aaa', 'bbb', 'c') == 'aaa'

    def test_min_mixed(self):
        assert min('1', 2, 3, 'aa') == 2

    def test_min_noargs(self):
        raises(TypeError, min)

    def test_min_empty(self):
        raises(ValueError, min, [])

class AppTestMax:

    def test_max_notseq(self):
        raises(TypeError, max, 1)

    def test_max_usual(self):
        assert max(1, 2, 3) == 3

    def test_max_floats(self):
        assert max(0.1, 2.7, 14.7) == 14.7

    def test_max_chars(self):
        assert max('a', 'b', 'c') == 'c'

    def test_max_strings(self):
        assert max('aaa', 'bbb', 'c') == 'c'

    def test_max_mixed(self):
        assert max('1', 2, 3, 'aa') == 'aa'

    def test_max_noargs(self):
        raises(TypeError, max)

    def test_max_empty(self):
        raises(ValueError, max, [])

class AppTestMaxTuple:

    def test_max_usual(self):
        assert max((1, 2, 3)) == 3

    def test_max_floats(self):
        assert max((0.1, 2.7, 14.7)) == 14.7

    def test_max_chars(self):
        assert max(('a', 'b', 'c')) == 'c'

    def test_max_strings(self):
        assert max(('aaa', 'bbb', 'c')) == 'c'

    def test_max_mixed(self):
        assert max(('1', 2, 3, 'aa')) == 'aa'

class AppTestMinList:

    def test_min_usual(self):
        assert min([1, 2, 3]) == 1

    def test_min_floats(self):
        assert min([0.1, 2.7, 14.7]) == 0.1

    def test_min_chars(self):
        assert min(['a', 'b', 'c']) == 'a'

    def test_min_strings(self):
        assert min(['aaa', 'bbb', 'c']) == 'aaa'

    def test_min_mixed(self):
        assert min(['1', 2, 3, 'aa']) == 2



from pypy.interpreter import module
from pypy.module.cpyext.api import (
    generic_cpy_call, cpython_api, PyObject, CONST_STRING, CANNOT_FAIL)
from rpython.rtyper.lltypesystem import lltype, rffi
from pypy.interpreter.error import OperationError
from pypy.interpreter.module import Module
from pypy.interpreter.pycode import PyCode
from pypy.module.imp import importing

@cpython_api([PyObject], PyObject)
def PyImport_Import(space, w_name):
    """
    This is a higher-level interface that calls the current "import hook function".
    It invokes the __import__() function from the __builtins__ of the
    current globals.  This means that the import is done using whatever import hooks
    are installed in the current environment, e.g. by rexec or ihooks.

    Always uses absolute imports."""
    caller = space.getexecutioncontext().gettopframe_nohidden()
    # Get the builtins from current globals
    if caller is not None:
        w_globals = caller.get_w_globals()
        w_builtin = space.getitem(w_globals, space.newtext('__builtins__'))
    else:
        # No globals -- use standard builtins, and fake globals
        w_builtin = space.getbuiltinmodule('__builtin__')
        w_globals = space.newdict()
        space.setitem(w_globals, space.newtext("__builtins__"), w_builtin)

    # Get the __import__ function from the builtins
    if space.isinstance_w(w_builtin, space.w_dict):
        w_import = space.getitem(w_builtin, space.newtext("__import__"))
    else:
        w_import = space.getattr(w_builtin, space.newtext("__import__"))

    # Call the __import__ function with the proper argument list
    # Always use absolute import here.
    return space.call_function(w_import,
                               w_name, w_globals, w_globals,
                               space.newlist([space.newtext("__doc__")]))

@cpython_api([CONST_STRING], PyObject)
def PyImport_ImportModule(space, name):
    return PyImport_Import(space, space.newtext(rffi.charp2str(name)))

@cpython_api([CONST_STRING], PyObject)
def PyImport_ImportModuleNoBlock(space, name):
    space.warn(
        space.newtext('PyImport_ImportModuleNoBlock() is not non-blocking'),
        space.w_RuntimeWarning)
    return PyImport_Import(space, space.newtext(rffi.charp2str(name)))

@cpython_api([PyObject], PyObject)
def PyImport_ReloadModule(space, w_mod):
    from pypy.module.imp.importing import reload
    return reload(space, w_mod)

@cpython_api([CONST_STRING], PyObject, result_borrowed=True)
def PyImport_AddModule(space, name):
    """Return the module object corresponding to a module name.  The name
    argument may be of the form package.module. First check the modules
    dictionary if there's one there, and if not, create a new one and insert
    it in the modules dictionary. Return NULL with an exception set on
    failure.

    This function does not load or import the module; if the module wasn't
    already loaded, you will get an empty module object. Use
    PyImport_ImportModule() or one of its variants to import a module.
    Package structures implied by a dotted name for name are not created if
    not already present."""
    from pypy.module.imp.importing import check_sys_modules_w
    modulename = rffi.charp2str(name)
    w_mod = check_sys_modules_w(space, modulename)
    if not w_mod or space.is_w(w_mod, space.w_None):
        w_mod = Module(space, space.newtext(modulename))
        space.setitem(space.sys.get('modules'), space.newtext(modulename), w_mod)
    # return a borrowed ref --- assumes one copy in sys.modules
    return w_mod

@cpython_api([], PyObject, result_borrowed=True)
def PyImport_GetModuleDict(space):
    """Return the dictionary used for the module administration (a.k.a.
    sys.modules).  Note that this is a per-interpreter variable."""
    w_modulesDict = space.sys.get('modules')
    return w_modulesDict     # borrowed ref

@cpython_api([rffi.CCHARP, PyObject], PyObject)
def PyImport_ExecCodeModule(space, name, w_code):
    """Given a module name (possibly of the form package.module) and a code
    object read from a Python bytecode file or obtained from the built-in
    function compile(), load the module.  Return a new reference to the module
    object, or NULL with an exception set if an error occurred.  Before Python
    2.4, the module could still be created in error cases.  Starting with Python
    2.4, name is removed from sys.modules in error cases, and even if name was
    already in sys.modules on entry to PyImport_ExecCodeModule().  Leaving
    incompletely initialized modules in sys.modules is dangerous, as imports of
    such modules have no way to know that the module object is an unknown (and
    probably damaged with respect to the module author's intents) state.

    The module's __file__ attribute will be set to the code object's
    co_filename.

    This function will reload the module if it was already imported.  See
    PyImport_ReloadModule() for the intended way to reload a module.

    If name points to a dotted name of the form package.module, any package
    structures not already created will still not be created.

    name is removed from sys.modules in error cases."""
    return PyImport_ExecCodeModuleEx(space, name, w_code,
                                     lltype.nullptr(rffi.CCHARP.TO))


@cpython_api([rffi.CCHARP, PyObject, rffi.CCHARP], PyObject)
def PyImport_ExecCodeModuleEx(space, name, w_code, pathname):
    """Like PyImport_ExecCodeModule(), but the __file__ attribute of
    the module object is set to pathname if it is non-NULL."""
    code = space.interp_w(PyCode, w_code)
    w_name = space.newtext(rffi.charp2str(name))
    if pathname:
        pathname = rffi.charp2str(pathname)
    else:
        pathname = code.co_filename
    w_mod = importing.add_module(space, w_name)
    space.setattr(w_mod, space.newtext('__file__'), space.newtext(pathname))
    return importing.exec_code_module(space, w_mod, code, w_name)

@cpython_api([], lltype.Void, error=CANNOT_FAIL)
def _PyImport_AcquireLock(space):
    """Locking primitive to prevent parallel imports of the same module
    in different threads to return with a partially loaded module.
    These calls are serialized by the global interpreter lock."""
    try:
        space.call_method(space.getbuiltinmodule('imp'), 'acquire_lock')
    except OperationError as e:
        e.write_unraisable(space, "_PyImport_AcquireLock")

@cpython_api([], rffi.INT_real, error=CANNOT_FAIL)
def _PyImport_ReleaseLock(space):
    try:
        space.call_method(space.getbuiltinmodule('imp'), 'release_lock')
        return 1
    except OperationError as e:
        e.write_unraisable(space, "_PyImport_ReleaseLock")
        return -1



import re
import os


for line in open('stubs.py'):
    if not line.strip():
        continue
    if line.startswith('    '):
        continue
    if line.startswith('#'):
        continue
    if line.startswith('@cpython_api'):
        continue
    if line.endswith(' = rffi.VOIDP\n'):
        continue

    #print line.rstrip()
    m = re.match(r"def ([\w\d_]+)[(]", line)
    assert m, line
    funcname = m.group(1)
    os.system('grep -w %s [a-r]*.py s[a-s]*.py str*.py stubsa*.py sy*.py [t-z]*.py' % funcname)



import contextlib

class Codebuilder(object):
    def __init__(self):
        self.blocks = []
        self.code = []

    def get_code(self):
        assert not self.blocks
        return "\n".join(["    " * depth + line for depth, line in self.code])

    def make_parser(self):
        m = {'Status': Status,
             'Nonterminal': Nonterminal,
             'Symbol': Symbol,}
        exec(py.code.Source(self.get_code()).compile(), m)
        return m['Parser']

    def emit(self, line):
        for line in line.split("\n"):
            if line:
                self.code.append((len(self.blocks),  line))

    def emit_initcode(self, line):
        for line in line.split("\n"):
            self.initcode.append(line)

    def start_block(self, blockstarter):
        assert blockstarter.endswith(":")
        self.emit(blockstarter)
        self.blocks.append(blockstarter)

    @contextlib.contextmanager
    def block(self, blockstarter):
        self.start_block(blockstarter)
        yield None
        self.end_block(blockstarter)

    def end_block(self, starterpart=""):
        block = self.blocks.pop()
        assert starterpart in block, "ended wrong block %s with %s" % (
            block, starterpart)

    def store_code_away(self):
        result = self.blocks, self.code
        self.code = []
        self.blocks = []
        return result

    def restore_code(self, (blocks, code)):
        result = self.blocks, self.code
        self.code = code
        self.blocks = blocks
        return result

    def add_code(self, (blocks, code)):
        self.code += [(depth + len(self.blocks), line) for depth, line in code]
        self.blocks += blocks
 





class AppTestSpecialTestCase:
    def test_Ellipsis(self):
        assert Ellipsis == Ellipsis
        assert repr(Ellipsis) == 'Ellipsis'
        assert Ellipsis.__class__.__name__ == 'ellipsis'
    
    def test_NotImplemented(self):
        def f():
            return NotImplemented
        assert f() == NotImplemented 
        assert repr(NotImplemented) == 'NotImplemented'
        assert NotImplemented.__class__.__name__ == 'NotImplementedType'



#!/usr/bin/env python

from rpython.jit.backend.llsupport import jitframe, rewrite
from rpython.jit.backend.llsupport.asmmemmgr import MachineDataBlockWrapper
from rpython.jit.backend.llsupport.descr import ArrayDescr
from rpython.jit.backend.model import CompiledLoopToken
from rpython.jit.backend.riscv import registers as r
from rpython.jit.backend.riscv.arch import (
    ABI_STACK_ALIGN, FLEN, INST_SIZE, JITFRAME_FIXED_SIZE,
    SCRATCH_STACK_SLOT_SIZE, XLEN)
from rpython.jit.backend.riscv.codebuilder import (
    InstrBuilder, MAX_NUM_INSTS_FOR_LOAD_INT_IMM, OverwritingBuilder)
from rpython.jit.backend.riscv.instruction_util import (
    can_fuse_into_compare_and_branch, check_imm_arg, check_simm21_arg)
from rpython.jit.backend.riscv.opassembler import (
    OpAssembler, asm_guard_operations, asm_operations)
from rpython.jit.backend.riscv.regalloc import (
    Regalloc, regalloc_guard_operations, regalloc_operations)
from rpython.jit.backend.riscv.locations import (
    ImmLocation, StackLocation, get_fp_offset)
from rpython.jit.codewriter.effectinfo import EffectInfo
from rpython.jit.metainterp.history import AbstractFailDescr, FLOAT
from rpython.jit.metainterp.resoperation import rop
from rpython.rlib import rgc, rmmap
from rpython.rlib.debug import debug_print, debug_start, debug_stop
from rpython.rlib.jit import AsmInfo
from rpython.rlib.objectmodel import compute_unique_id, we_are_translated
from rpython.rlib.rarithmetic import r_uint
from rpython.rlib.rjitlog import rjitlog as jl
from rpython.rtyper.annlowlevel import cast_instance_to_gcref
from rpython.rtyper.lltypesystem import lltype, rffi


# Maximum size for an absolute branch stub (load addr + jalr).
_REDIRECT_BRANCH_STUB_SIZE = (MAX_NUM_INSTS_FOR_LOAD_INT_IMM + 1) * INST_SIZE

def _emit_nop_until_larger(mc, start, end):
    for i in range(start, end, INST_SIZE):
        mc.NOP()

class AssemblerRISCV(OpAssembler):
    def __init__(self, cpu, translate_support_code=False):
        OpAssembler.__init__(self, cpu, translate_support_code)
        self.failure_recovery_code = [0, 0, 0, 0]
        self.propagate_exception_path = 0
        self.wb_slowpath = [0, 0, 0, 0, 0]
        self.stack_check_slowpath = 0
        self._frame_realloc_slowpath = 0

    @rgc.no_release_gil
    def assemble_loop(self, jd_id, unique_id, logger, loopname, inputargs,
                      operations, looptoken, log):
        rmmap.enter_assembler_writing()
        try:
            return self._assemble_loop(jd_id, unique_id, logger, loopname,
                                       inputargs, operations, looptoken, log)
        finally:
            rmmap.leave_assembler_writing()

    def _assemble_loop(self, jd_id, unique_id, logger, loopname, inputargs,
                       operations, looptoken, log):
        if not we_are_translated():
            # Arguments should be unique
            assert len(set(inputargs)) == len(inputargs)

        clt = CompiledLoopToken(self.cpu, looptoken.number)
        clt._debug_nbargs = len(inputargs)
        looptoken.compiled_loop_token = clt

        self.setup(looptoken)
        if self.cpu.HAS_CODEMAP:
            self.codemap_builder.enter_portal_frame(jd_id, unique_id,
                                                    self.mc.get_relative_pos())

        frame_info = self.datablockwrapper.malloc_aligned(
            jitframe.JITFRAMEINFO_SIZE, alignment=XLEN)
        clt.frame_info = rffi.cast(jitframe.JITFRAMEINFOPTR, frame_info)
        clt.frame_info.clear()

        if log:
            operations = self._inject_debugging_code(looptoken, operations,
                                                     'e', looptoken.number)

        regalloc = Regalloc(self)
        allgcrefs = []
        operations = regalloc.prepare_loop(inputargs, operations, looptoken,
                                           allgcrefs)
        self.reserve_gcref_table(allgcrefs)
        function_pos = self.mc.get_relative_pos()

        self._call_header_with_stack_check()

        loop_head = self.mc.get_relative_pos()
        looptoken._ll_loop_code = loop_head

        frame_depth_no_fixed_size = self._assemble(regalloc, inputargs,
                                                   operations)
        frame_depth = frame_depth_no_fixed_size + JITFRAME_FIXED_SIZE
        self.update_frame_depth(frame_depth)
        self.patch_frame_depth_checks(frame_depth)

        # Generate extra NOPs if the size is too small. We need this because
        # `redirect_call_assembler` may want to patch the beginning with a far
        # branch to another loop or bridge.
        _emit_nop_until_larger(self.mc, self.mc.get_relative_pos(),
                               function_pos + _REDIRECT_BRANCH_STUB_SIZE)

        size_excluding_failure_stuff = self.mc.get_relative_pos()

        self.write_pending_failure_recoveries()

        const_pos = self.mc.get_relative_pos()
        self.mc.emit_pending_constants()

        full_size = self.mc.get_relative_pos()
        rawstart = self.materialize_loop(looptoken)
        looptoken._ll_function_addr = rawstart + function_pos

        self.patch_gcref_table(looptoken, rawstart)
        self.process_pending_guards(rawstart)
        self.fixup_target_tokens(rawstart)

        if log and not we_are_translated():
            self.mc._dump_trace(rawstart, 'loop.asm')

        ops_offset = self.mc.ops_offset

        if logger:
            log = logger.log_trace(jl.MARK_TRACE_ASM, None, self.mc)
            log.write(inputargs, operations, ops_offset=ops_offset)

            if logger.logger_ops:
                logger.logger_ops.log_loop(inputargs, operations, 0,
                                           'rewritten', name=loopname,
                                           ops_offset=ops_offset)

        debug_start('jit-backend-addr')
        debug_print('Loop %d (%s) has address 0x%x to 0x%x (bootstrap 0x%x)' % (
            looptoken.number, loopname,
            r_uint(rawstart + loop_head),
            r_uint(rawstart + size_excluding_failure_stuff),
            r_uint(rawstart + function_pos)))
        debug_print('       gc table: 0x%x' % r_uint(rawstart))
        debug_print('       function: 0x%x' % r_uint(rawstart + function_pos))
        debug_print('         resops: 0x%x' % r_uint(rawstart + loop_head))
        debug_print('       failures: 0x%x' % r_uint(rawstart +
                                                 size_excluding_failure_stuff))
        debug_print("     const pool: 0x%x" % r_uint(rawstart + const_pos))
        debug_print('            end: 0x%x' % r_uint(rawstart + full_size))
        debug_stop('jit-backend-addr')

        self.teardown()

        return AsmInfo(ops_offset, rawstart + loop_head,
                       size_excluding_failure_stuff - loop_head)

    @rgc.no_release_gil
    def assemble_bridge(self, logger, faildescr, inputargs, operations,
                        original_loop_token, log):
        rmmap.enter_assembler_writing()
        try:
            return self._assemble_bridge(logger, faildescr, inputargs,
                                         operations, original_loop_token, log)
        finally:
            rmmap.leave_assembler_writing()

    def _assemble_bridge(self, logger, faildescr, inputargs, operations,
                         original_loop_token, log):
        if not we_are_translated():
            # Arguments should be unique
            assert len(set(inputargs)) == len(inputargs)

        self.setup(original_loop_token)
        if self.cpu.HAS_CODEMAP:
            self.codemap_builder.inherit_code_from_position(
                faildescr.adr_jump_offset)

        descr_number = compute_unique_id(faildescr)
        if log:
            operations = self._inject_debugging_code(faildescr, operations,
                                                     'b', descr_number)

        assert isinstance(faildescr, AbstractFailDescr)

        arglocs = self.rebuild_faillocs_from_descr(faildescr, inputargs)

        regalloc = Regalloc(self)
        allgcrefs = []
        operations = regalloc.prepare_bridge(inputargs, arglocs, operations,
                                             allgcrefs,
                                             self.current_clt.frame_info)
        self.reserve_gcref_table(allgcrefs)
        start_pos = self.mc.get_relative_pos()

        self._check_frame_depth(self.mc, regalloc.get_gcmap(),
                                expected_size=-1)

        bridge_start_pos = self.mc.get_relative_pos()
        frame_depth_no_fixed_size = self._assemble(regalloc, inputargs,
                                                   operations)

        # Patch frame depth check.
        bridge_frame_depth = frame_depth_no_fixed_size + JITFRAME_FIXED_SIZE
        self.patch_frame_depth_checks(bridge_frame_depth)

        # Generate extra NOPs if the size is too small. We need this because
        # `redirect_call_assembler` may want to patch the beginning with a far
        # branch to another loop or bridge.
        _emit_nop_until_larger(self.mc, self.mc.get_relative_pos(),
                               start_pos + _REDIRECT_BRANCH_STUB_SIZE)

        code_end_pos = self.mc.get_relative_pos()

        self.write_pending_failure_recoveries()

        const_pos = self.mc.get_relative_pos()
        self.mc.emit_pending_constants()

        fullsize = self.mc.get_relative_pos()
        rawstart = self.materialize_loop(original_loop_token)

        self.patch_gcref_table(original_loop_token, rawstart)
        self.process_pending_guards(rawstart)

        # Update the frame depth in compiled_loop_token.
        frame_depth = max(self.current_clt.frame_info.jfi_frame_depth,
                          bridge_frame_depth)
        self.update_frame_depth(frame_depth)

        # Replace _ll_loop_code relative offset with an absolute address.
        self.fixup_target_tokens(rawstart)

        # Patch the jump from original guard.
        self.patch_trace(faildescr, original_loop_token, rawstart + start_pos,
                         regalloc)

        if log and not we_are_translated():
            self.mc._dump_trace(rawstart, 'bridge.asm')

        ops_offset = self.mc.ops_offset

        if logger:
            log = logger.log_trace(jl.MARK_TRACE_ASM, None, self.mc)
            log.write(inputargs, operations, ops_offset)
            # Log that the already written bridge is stitched to a descr.
            logger.log_patch_guard(descr_number, rawstart)

            # Legacy
            if logger.logger_ops:
                logger.logger_ops.log_bridge(inputargs, operations,
                                             'rewritten', faildescr,
                                             ops_offset=ops_offset)

        debug_start("jit-backend-addr")
        debug_print("bridge out of Guard 0x%x has address 0x%x to 0x%x" %
                    (r_uint(descr_number), r_uint(rawstart + start_pos),
                        r_uint(rawstart + code_end_pos)))
        debug_print("       gc table: 0x%x" % r_uint(rawstart))
        debug_print("    jump target: 0x%x" % r_uint(rawstart + start_pos))
        debug_print("         resops: 0x%x" % r_uint(rawstart +
                                                     bridge_start_pos))
        debug_print("       failures: 0x%x" % r_uint(rawstart + code_end_pos))
        debug_print("     const pool: 0x%x" % r_uint(rawstart + const_pos))
        debug_print("            end: 0x%x" % r_uint(rawstart + fullsize))
        debug_stop("jit-backend-addr")

        self.teardown()

        return AsmInfo(ops_offset, start_pos + rawstart,
                       code_end_pos - start_pos)

    @rgc.no_release_gil
    def redirect_call_assembler(self, oldlooptoken, newlooptoken):
        rmmap.enter_assembler_writing()
        try:
            self._redirect_call_assembler(oldlooptoken, newlooptoken)
        finally:
            rmmap.leave_assembler_writing()

    def _redirect_call_assembler(self, oldlooptoken, newlooptoken):
        # Some minimal sanity checking
        old_nbargs = oldlooptoken.compiled_loop_token._debug_nbargs
        new_nbargs = newlooptoken.compiled_loop_token._debug_nbargs
        assert old_nbargs == new_nbargs

        # We overwrite the instructions at the old `_ll_function_addr` to start
        # with a jump to the new _ll_function_addr.
        oldadr = oldlooptoken._ll_function_addr
        target = newlooptoken._ll_function_addr

        # Copy frame info
        baseofs = self.cpu.get_baseofs_of_frame_field()
        newlooptoken.compiled_loop_token.update_frame_info(
            oldlooptoken.compiled_loop_token, baseofs)

        mc = InstrBuilder()
        scratch_reg = r.x31  # Pick a caller-saved reg excluding x10-17 & ra
        mc.load_int_imm(scratch_reg.value, target)
        mc.JALR(r.x0.value, scratch_reg.value, 0)
        mc.emit_pending_constants()
        mc.copy_to_raw_memory(oldadr)

        jl.redirect_assembler(oldlooptoken, newlooptoken, newlooptoken.number)

    def _assemble(self, regalloc, inputargs, operations):
        # Fill in the frame location hints so that we can reduce stack-to-stack
        # data movement in `remap_frame_layout_mixed`.
        regalloc.compute_hint_frame_locations(operations)

        # Visit all operations and regalloc/assemble the operations.
        self._walk_operations(inputargs, operations, regalloc)
        frame_depth = regalloc.get_final_frame_depth()

        # If the jump target (of the current loop) requires larger frame,
        # update the frame depth.
        jump_target_descr = regalloc.jump_target_descr
        if jump_target_descr is not None:
            tgt_depth = jump_target_descr._riscv_clt.frame_info.jfi_frame_depth
            target_frame_depth = tgt_depth - JITFRAME_FIXED_SIZE
            frame_depth = max(frame_depth, target_frame_depth)

        return frame_depth

    def _walk_operations(self, inputargs, operations, regalloc):
        self._regalloc = regalloc
        regalloc.operations = operations

        while regalloc.position() < len(operations) - 1:
            regalloc.next_instruction()
            i = regalloc.position()
            op = operations[i]
            self.mc.mark_op(op)
            opnum = op.getopnum()

            if rop.has_no_side_effect(opnum) and op not in regalloc.longevity:
                # If this op does not have side effects and its result is
                # unused, it is safe to ignore this op.
                pass
            elif not we_are_translated() and op.getopnum() == rop.FORCE_SPILL:
                regalloc.force_spill_var(op.getarg(0))
            elif (i < len(operations) - 1 and
                  ((can_fuse_into_compare_and_branch(opnum) and
                    regalloc.next_op_can_accept_cc(operations, i)) or
                   (op.is_ovf() and
                    rop.is_guard_overflow(operations[i + 1].getopnum())))):
                guard_op = operations[i + 1]  # guard_* or cond_call*
                guard_num = guard_op.getopnum()
                arglocs, guard_branch_inst = \
                        regalloc_guard_operations[guard_num](regalloc, op,
                                                             guard_op)
                if arglocs is not None:
                    asm_guard_operations[guard_num](self, op, guard_op, arglocs,
                                                    guard_branch_inst)
                regalloc.next_instruction()  # Advance one more
                # Free argument vars of the guard op (if no longer used).
                if guard_op.is_guard():
                    regalloc.possibly_free_vars(guard_op.getfailargs())
                regalloc.possibly_free_vars_for_op(guard_op)
                # Free the return var of the guard op (if no longer used).
                regalloc.possibly_free_var(guard_op)
            elif (rop.is_call_may_force(op.getopnum()) or
                  rop.is_call_release_gil(op.getopnum()) or
                  rop.is_call_assembler(op.getopnum())):
                guard_op = operations[i + 1]
                guard_num = guard_op.getopnum()
                assert guard_num in (rop.GUARD_NOT_FORCED,
                                     rop.GUARD_NOT_FORCED_2)

                # `arglocs` contains the locations for `op` and `guard_op`.
                # The first `num_arglocs` locations are for `op` and the
                # remainings are for `guard_op`.
                arglocs, num_arglocs = \
                        regalloc_guard_operations[guard_num](regalloc, op,
                                                             guard_op)
                if arglocs is not None:
                    asm_guard_operations[guard_num](self, op, guard_op, arglocs,
                                                    num_arglocs)
                regalloc.next_instruction()  # Advance one more

                # Free argument vars of the guard op (if no longer used).
                regalloc.possibly_free_vars(guard_op.getfailargs())
                regalloc.possibly_free_vars_for_op(guard_op)
            else:
                arglocs = regalloc_operations[opnum](regalloc, op)
                if arglocs is not None:
                    asm_operations[opnum](self, op, arglocs)

            # Free argument vars of the op (if no longer used).
            regalloc.possibly_free_vars_for_op(op)
            if rop.is_guard(opnum):
                regalloc.possibly_free_vars(op.getfailargs())

            # Free the return var of the op (if no longer used).
            #
            # Note: This can happen when we want the side-effect of an op (e.g.
            # `call_assembler_i` or `call_i`) but want to discard the returned
            # value.
            if op.type != 'v':
                regalloc.possibly_free_var(op)

            regalloc.free_temp_vars()
            regalloc._check_invariants()

        if not we_are_translated():
            self.mc.EBREAK()
        self.mc.mark_op(None)  # End of the loop
        regalloc.operations = None

    def _call_header_with_stack_check(self):
        self._call_header()

        if self.stack_check_slowpath == 0:
            pass  # No stack check (e.g. not translated)
        else:
            endaddr, lengthaddr, _ = self.cpu.insert_stack_check()

            # hi: endaddr
            #     stack pointer
            # lo: startaddr = endaddr - lengthaddr

            scratch_reg = r.x31
            scratch2_reg = r.x30

            # Load stack end
            self.mc.load_int_imm(scratch_reg.value, endaddr)
            self.mc.load_int(scratch_reg.value, scratch_reg.value, 0)

            # Load stack length
            self.mc.load_int_imm(scratch2_reg.value, lengthaddr)
            self.mc.load_int(scratch2_reg.value, scratch2_reg.value, 0)

            # Calculate stack_start = stack_end - stack_len
            self.mc.SUB(scratch_reg.value, scratch_reg.value,
                        scratch2_reg.value)

            # Patch Location: BGEU sp, stack_start, end
            pos = self.mc.get_relative_pos()
            self.mc.EBREAK()

            self.mc.load_int_imm(r.ra.value, self.stack_check_slowpath)
            self.mc.JALR(r.ra.value, r.ra.value, 0)

            # LABEL[end]:
            offset = self.mc.get_relative_pos() - pos
            pmc = OverwritingBuilder(self.mc, pos, INST_SIZE)
            pmc.BGEU(r.sp.value, scratch_reg.value, offset)

    def _call_header(self):
        self._push_callee_save_regs_to_stack(self.mc)

        if self.cpu.translate_support_code:
            self._call_header_vmprof()

        # Save the thread local address to tls[0].
        self.saved_threadlocal_addr = 0 * XLEN
        self.mc.store_int(r.x11.value, r.sp.value, 0 * XLEN)

        self.mc.MV(r.jfp.value, r.x10.value)

        gcrootmap = self.cpu.gc_ll_descr.gcrootmap
        if gcrootmap and gcrootmap.is_shadow_stack:
            self.gen_shadowstack_header(gcrootmap)

    def _call_footer(self, mc):
        gcrootmap = self.cpu.gc_ll_descr.gcrootmap
        if gcrootmap and gcrootmap.is_shadow_stack:
            self.gen_shadowstack_footer(gcrootmap, mc)

        if self.cpu.translate_support_code:
            self._call_footer_vmprof(mc)

        mc.MV(r.x10.value, r.jfp.value)
        self._pop_callee_save_regs_from_stack(mc)
        mc.RET()

    def gen_shadowstack_header(self, gcrootmap):
        scratch_reg = r.x31
        scratch2_reg = r.x30
        scratch3_reg = r.x29

        # scratch_reg = &root_stack_top_addr (address to pointer to stack top)
        rst = gcrootmap.get_root_stack_top_addr()
        self.mc.load_int_imm(scratch_reg.value, rst)
        # scratch2_reg = root_stack_top_addr (address to stack top)
        self.mc.load_int(scratch2_reg.value, scratch_reg.value, 0)

        # We push two words, like the x86 backend does:
        # The '1' is to benefit from the shadowstack 'is_minor' optimization

        # scratch2_reg[0] = 1
        self.mc.load_int_imm(scratch3_reg.value, 1)
        self.mc.store_int(scratch3_reg.value, scratch2_reg.value, 0)
        # scratch2_reg[1] = r.jfp
        self.mc.store_int(r.jfp.value, scratch2_reg.value, XLEN)

        # scratch2_reg += 2 * XLEN
        self.mc.ADDI(scratch2_reg.value, scratch2_reg.value, 2 * XLEN)
        # root_stack_top_addr = scratch2_reg
        self.mc.store_int(scratch2_reg.value, scratch_reg.value, 0)

    def gen_shadowstack_footer(self, gcrootmap, mc):
        scratch_reg = r.x31
        scratch2_reg = r.x30

        # scratch_reg = &root_stack_top_addr
        rst = gcrootmap.get_root_stack_top_addr()
        mc.load_int_imm(scratch_reg.value, rst)
        # scratch2_reg = root_stack_top_addr
        mc.load_int(scratch2_reg.value, scratch_reg.value, 0)

        # scratch2_reg -= 2 * XLEN
        mc.ADDI(scratch2_reg.value, scratch2_reg.value, -(2 * XLEN))
        # root_stack_top_addr = scratch2_reg
        mc.store_int(scratch2_reg.value, scratch_reg.value, 0)

    def _call_header_vmprof(self):
        from rpython.rlib.rvmprof.rvmprof import cintf, VMPROF_JITTED_TAG

        # tloc = &pypy_threadlocal_s
        tloc = r.x11
        scratch_reg = r.x31
        scratch2_reg = r.x30

        # scratch_reg = &vmprof_tl_stack (old vmprof_tl_stack top)
        offset = rffi.cast(lltype.Signed, cintf.vmprof_tl_stack.getoffset())
        self.mc.load_int_from_base_plus_offset(scratch_reg.value, tloc.value,
                                               offset)
        # stack->next = old vmprof_tl_stack top
        self.mc.store_int(scratch_reg.value, r.sp.value, 1 * XLEN)
        # stack->value = sp
        self.mc.store_int(r.sp.value, r.sp.value, 2 * XLEN)
        # stack->kind = VMPROF_JITTED_TAG
        self.mc.load_int_imm(scratch_reg.value, VMPROF_JITTED_TAG)
        self.mc.store_int(scratch_reg.value, r.sp.value, 3 * XLEN)
        # Set vmprof_tl_stack top to the new entry.
        self.mc.ADDI(scratch_reg.value, r.sp.value, 1 * XLEN)
        self.mc.store_int_to_base_plus_offset(scratch_reg.value, tloc.value,
                                              offset, tmp=scratch2_reg.value)

    def _call_footer_vmprof(self, mc):
        from rpython.rlib.rvmprof.rvmprof import cintf

        tloc = r.x11
        scratch_reg = r.x31
        scratch2_reg = r.x30

        # tloc = &pypy_threadlocal_s
        mc.load_int(tloc.value, r.sp.value, 0 * XLEN)
        # scratch_reg = thread local vmprof_tl_stack->next
        mc.load_int(scratch_reg.value, r.sp.value, 1 * XLEN)
        # Set vmprof_tl_stack top to vmprof_tl_stack->next (pop stack)
        offset = rffi.cast(lltype.Signed, cintf.vmprof_tl_stack.getoffset())
        mc.store_int_to_base_plus_offset(scratch_reg.value, tloc.value,
                                         offset, tmp=scratch2_reg.value)

    def _calculate_callee_save_area_size(self):
        # Extra thread local storage.
        #
        # tls[0 * XLEN]: saved_threadlocal_addr (_call_header)
        # tls[1 * XLEN]: VMPROFSTACK->next  (vmprof)
        # tls[2 * XLEN]: VMPROFSTACK->value (vmprof)
        # tls[3 * XLEN]: VMPROFSTACK->kind  (vmprof)
        tls_size = XLEN * 4

        core_reg_begin = tls_size
        core_reg_size = XLEN * len(r.callee_saved_registers_except_ra_sp_fp)

        fp_reg_begin = core_reg_begin + core_reg_size
        fp_reg_begin = (fp_reg_begin + FLEN - 1) // FLEN * FLEN
        fp_reg_size = FLEN * len(r.callee_saved_fp_registers)

        # fp = old_sp
        # frame_record[0 * XLEN] (or fp[-2 * XLEN]): fp (old)
        # frame_record[1 * XLEN] (or fp[-1 * XLEN]): ra
        frame_record_begin = fp_reg_begin + fp_reg_size
        frame_record_begin = (frame_record_begin + XLEN - 1) // XLEN * XLEN
        frame_record_size = 2 * XLEN

        area_size = frame_record_begin + frame_record_size
        area_size = ((area_size + ABI_STACK_ALIGN - 1)
                         // ABI_STACK_ALIGN * ABI_STACK_ALIGN)

        frame_record_begin = area_size - frame_record_size

        return area_size, core_reg_begin, fp_reg_begin, frame_record_begin

    def _push_callee_save_regs_to_stack(self, mc):
        area_size, core_reg_begin, fp_reg_begin, frame_record_begin = \
                self._calculate_callee_save_area_size()

        # Subtract stack pointer
        mc.ADDI(r.sp.value, r.sp.value, -area_size)

        # Frame record
        mc.store_int(r.fp.value, r.sp.value, frame_record_begin)
        mc.store_int(r.ra.value, r.sp.value, frame_record_begin + XLEN)
        mc.ADDI(r.fp.value, r.sp.value, area_size)

        for i, reg in enumerate(r.callee_saved_registers_except_ra_sp_fp):
            mc.store_int(reg.value, r.sp.value, i * XLEN + core_reg_begin)
        for i, reg in enumerate(r.callee_saved_fp_registers):
            mc.store_float(reg.value, r.sp.value, i * FLEN + fp_reg_begin)

    def _pop_callee_save_regs_from_stack(self, mc):
        area_size, core_reg_begin, fp_reg_begin, frame_record_begin = \
                self._calculate_callee_save_area_size()
        for i, reg in enumerate(r.callee_saved_fp_registers):
            mc.load_float(reg.value, r.sp.value, i * FLEN + fp_reg_begin)
        for i, reg in enumerate(r.callee_saved_registers_except_ra_sp_fp):
            mc.load_int(reg.value, r.sp.value, i * XLEN + core_reg_begin)

        # Frame record
        mc.load_int(r.ra.value, r.sp.value, frame_record_begin + XLEN)
        mc.load_int(r.fp.value, r.sp.value, frame_record_begin)

        # Add (restore) stack pointer
        mc.ADDI(r.sp.value, r.sp.value, area_size)

    def _push_all_regs_to_jitframe(self, mc, ignored_regs, withfloats,
                                   callee_only=False):
        # Push general purpose registers
        base_ofs = self.cpu.get_baseofs_of_frame_field()
        if callee_only:
            regs = r.caller_saved_registers
        else:
            regs = r.registers_except_zero

        if not ignored_regs:
            for reg in regs:
                mc.store_int(reg.value, r.jfp.value,
                             base_ofs +
                             self.cpu.all_reg_indexes[reg.value] * XLEN)
        else:
            for reg in ignored_regs:
                assert reg.is_core_reg()
            for reg in regs:
                if reg in ignored_regs:
                    continue
                mc.store_int(reg.value, r.jfp.value,
                             base_ofs +
                             self.cpu.all_reg_indexes[reg.value] * XLEN)

        if withfloats:
            # Push floating point registers
            ofs = base_ofs + len(r.registers) * XLEN
            for reg in r.fp_registers:
                mc.store_float(reg.value, r.jfp.value, ofs + reg.value * FLEN)

    def _pop_all_regs_from_jitframe(self, mc, ignored_regs, withfloats,
                                    callee_only=False):
        # Pop general purpose registers
        base_ofs = self.cpu.get_baseofs_of_frame_field()
        if callee_only:
            regs = r.caller_saved_registers
        else:
            regs = r.registers_except_zero

        if not ignored_regs:
            for reg in regs:
                mc.load_int(reg.value, r.jfp.value,
                            base_ofs +
                            self.cpu.all_reg_indexes[reg.value] * XLEN)
        else:
            for reg in ignored_regs:
                assert reg.is_core_reg()
            for reg in regs:
                if reg in ignored_regs:
                    continue
                mc.load_int(reg.value, r.jfp.value,
                            base_ofs +
                            self.cpu.all_reg_indexes[reg.value] * XLEN)

        if withfloats:
            # Pop floating point registers
            ofs = base_ofs + len(r.registers) * XLEN
            for reg in r.fp_registers:
                mc.load_float(reg.value, r.jfp.value, ofs + reg.value * FLEN)

    def _push_regs_to_jitframe(self, mc, selected_regs):
        # Push specified regs to JITFrame.
        base_ofs = self.cpu.get_baseofs_of_frame_field()
        fp_ofs = base_ofs + len(r.registers) * XLEN
        for reg in selected_regs:
            if reg.is_core_reg():
                mc.store_int(reg.value, r.jfp.value,
                             base_ofs +
                             self.cpu.all_reg_indexes[reg.value] * XLEN)
            else:
                assert reg.is_fp_reg()
                mc.store_float(reg.value, r.jfp.value,
                               fp_ofs + reg.value * FLEN)

    def _pop_regs_from_jitframe(self, mc, selected_regs):
        # Pop specified regs from JITFrame.
        base_ofs = self.cpu.get_baseofs_of_frame_field()
        fp_ofs = base_ofs + len(r.registers) * XLEN
        for reg in selected_regs:
            if reg.is_core_reg():
                mc.load_int(reg.value, r.jfp.value,
                            base_ofs +
                            self.cpu.all_reg_indexes[reg.value] * XLEN)
            else:
                assert reg.is_fp_reg()
                mc.load_float(reg.value, r.jfp.value,
                              fp_ofs + reg.value * FLEN)

    def store_jf_descr(self, descrindex):
        scratch_reg = r.x31
        ofs = self.cpu.get_ofs_of_frame_field('jf_descr')
        self.load_from_gc_table(scratch_reg.value, descrindex)
        self.mc.store_int(scratch_reg.value, r.jfp.value, ofs)

    def push_gcmap(self, mc, gcmap, store=True):
        # Set gcmap address to jf_gcmap field.

        # rpython/jit/backend/llsupport/callbuilder.py passes a `store`
        # argument as keyword args. For RISC-V backend, we only support
        # `store=True` version.
        assert store

        scratch_reg = r.x31
        new_gcmap_adr = rffi.cast(lltype.Signed, gcmap)
        mc.load_int_imm(scratch_reg.value, new_gcmap_adr)

        ofs = self.cpu.get_ofs_of_frame_field('jf_gcmap')
        mc.store_int(scratch_reg.value, r.jfp.value, ofs)

    def pop_gcmap(self, mc):
        # Clear gcmap address from jf_gcmap field.
        ofs = self.cpu.get_ofs_of_frame_field('jf_gcmap')
        mc.store_int(r.x0.value, r.jfp.value, ofs)

    def patch_trace(self, faildescr, looptoken, bridge_addr, regalloc):
        # Patch the quick failure stub to jump to the bridge.

        # Before:
        #
        #     old_trace:
        #
        #         ... instructions  ...
        #
        #         if guard_fails:
        #             goto quick_failure_stub_i
        #
        #         ... instructions  ...
        #
        #         quick_failure_stub_i:
        #             push_gcmap()
        #             goto failure_recovery_code
        #
        # After:
        #
        #     old_trace:
        #
        #         ... instructions  ...
        #
        #         if guard_fails:
        #             goto quick_failure_stub_i
        #
        #         ... instructions  ...
        #
        #         quick_failure_stub_i:
        #             bridge_addr = ...
        #             goto bridge_addr

        patch_addr = faildescr.adr_jump_offset
        assert patch_addr != 0

        pmc = InstrBuilder()
        pmc.load_int_imm(r.x31.value, bridge_addr)
        pmc.JR(r.x31.value)
        pmc.emit_pending_constants()
        pmc.copy_to_raw_memory(patch_addr)

        faildescr.adr_jump_offset = 0

    def generate_quick_failure(self, guardtok):
        startpos = self.mc.get_relative_pos()
        faildescrindex, target = self.store_info_on_descr(startpos, guardtok)

        self.store_jf_descr(faildescrindex)
        self.push_gcmap(self.mc, guardtok.gcmap)
        assert target
        self.mc.jal_abs(r.zero.value, target)

        # Generate extra NOPs if this stub size is too small. We need this
        # padding because `patch_trace` will patch this stub to jump to the
        # compiled bridge.
        _emit_nop_until_larger(self.mc, self.mc.get_relative_pos(),
                               startpos + _REDIRECT_BRANCH_STUB_SIZE)
        return startpos

    def write_pending_failure_recoveries(self):
        for guardtok in self.pending_guards:
            guardtok.pos_recovery_stub = self.generate_quick_failure(guardtok)

    def process_pending_guards(self, rawstart):
        clt = self.current_clt
        for guardtok in self.pending_guards:
            descr = guardtok.faildescr
            assert isinstance(descr, AbstractFailDescr)

            failure_recovery_pos = rawstart + guardtok.pos_recovery_stub
            descr.adr_jump_offset = failure_recovery_pos
            relative_offset = guardtok.pos_recovery_stub - guardtok.offset
            guard_pos = rawstart + guardtok.offset

            if guardtok.guard_not_invalidated():
                clt.invalidate_positions.append((guard_pos, relative_offset))
            else:
                # Patch the guard jump to the stub
                assert check_simm21_arg(relative_offset)
                mc = InstrBuilder()
                mc.J(relative_offset)
                mc.copy_to_raw_memory(guard_pos)

    def fixup_target_tokens(self, rawstart):
        for targettoken in self.target_tokens_currently_compiling:
            targettoken._ll_loop_code += rawstart
        self.target_tokens_currently_compiling = None

    def reserve_gcref_table(self, allgcrefs):
        gcref_table_size = len(allgcrefs) * XLEN
        gcref_table_size = (gcref_table_size + 15) & ~15  # Align to 16

        # Reserve space at the beginning of the machine code for the gc table.
        # This lets us access gc table with pc-relative addressing.
        mc = self.mc
        assert mc.get_relative_pos() == 0
        for i in range(gcref_table_size):
            mc.writechar('\x00')

        self.setup_gcrefs_list(allgcrefs)

    def patch_gcref_table(self, looptoken, rawstart):
        self.gc_table_addr = rawstart
        tracer = self.cpu.gc_ll_descr.make_gcref_tracer(rawstart,
                                                        self._allgcrefs)
        gcreftracers = self.get_asmmemmgr_gcreftracers(looptoken)
        gcreftracers.append(tracer)  # Keepalive
        self.teardown_gcrefs_list()

    def load_from_gc_table(self, reg_num, index):
        address_in_buffer = index * XLEN  # at the start of the buffer
        p_location = self.mc.get_relative_pos(break_basic_block=False)
        offset = address_in_buffer - p_location
        self.mc.load_int_pc_rel(reg_num, offset)

    def setup(self, looptoken):
        OpAssembler.setup(self, looptoken)
        assert self.memcpy_addr != 0, 'setup_once() not called?'

        self.current_clt = looptoken.compiled_loop_token
        self.mc = InstrBuilder()
        self.pending_guards = []
        self.target_tokens_currently_compiling = {}

        allblocks = self.get_asmmemmgr_blocks(looptoken)
        self.datablockwrapper = MachineDataBlockWrapper(self.cpu.asmmemmgr,
                                                        allblocks)
        self.mc.datablockwrapper = self.datablockwrapper

        self._frame_depth_to_patch = []
        self._finish_gcmap = jitframe.NULLGCMAP

    def teardown(self):
        self.current_clt = None
        self._regalloc = None
        self.mc = None
        self.pending_guards = None

    def materialize_loop(self, looptoken):
        # Finalizes data block
        self.datablockwrapper.done()
        self.datablockwrapper = None

        # Finalizes instruction builder, combines the code buffers, and copy
        # them to an executable memory region.
        allblocks = self.get_asmmemmgr_blocks(looptoken)
        size = self.mc.get_relative_pos()
        rawstart = self.mc.materialize(self.cpu, allblocks,
                                       self.cpu.gc_ll_descr.gcrootmap)
        # Registers the materialized loop to the codemap.
        self.cpu.codemap.register_codemap(
            self.codemap_builder.get_final_bytecode(rawstart, size))
        return rawstart

    def _build_failure_recovery(self, exc, withfloats=False):
        mc = InstrBuilder()
        self._push_all_regs_to_jitframe(mc, [], withfloats)

        if exc:
            # Move the exception from `self.cpu.pos_exc_value()` to JITFrame
            # `jf_guard_exc` and then reset the data in
            # `self.cpu.pos_exc_value()` and `self.cpu.pos_exception()`.

            scratch_reg = r.x31
            scratch2_reg = r.x10  # Will be set by `_call_footer` soon.

            # Load exc_value from `self.cpu.pos_exc_value()`.
            mc.load_int_imm(scratch_reg.value, self.cpu.pos_exc_value())
            mc.load_int(scratch2_reg.value, scratch_reg.value, 0)

            # Clear `self.cpu.pos_exc_value()`.
            mc.store_int(r.x0.value, scratch_reg.value, 0)

            # Store exc_value to `jf_guard_exc`.
            ofs = self.cpu.get_ofs_of_frame_field('jf_guard_exc')
            mc.store_int(scratch2_reg.value, r.jfp.value, ofs)

            # Clear `self.cpu.pos_exception()`.
            mc.load_int_imm(scratch_reg.value, self.cpu.pos_exception())
            mc.store_int(r.x0.value, scratch_reg.value, 0)

        self._call_footer(mc)
        mc.emit_pending_constants()

        rawstart = mc.materialize(self.cpu, [])
        self.failure_recovery_code[exc + 2 * withfloats] = rawstart

    def propagate_memoryerror_if_reg_is_null(self, reg_loc):
        # Patch Location: BNEZ reg_loc, end
        cond_branch_addr = self.mc.get_relative_pos()
        self.mc.EBREAK()

        # Branch to `propagate_exception_path`
        self.mc.load_int_imm(r.ra.value, self.propagate_exception_path)
        self.mc.JR(r.ra.value)

        # LABEL[end]:
        offset = self.mc.get_relative_pos() - cond_branch_addr
        pmc = OverwritingBuilder(self.mc, cond_branch_addr, INST_SIZE)
        pmc.BNEZ(reg_loc.value, offset)

    def _build_wb_slowpath(self, withcards, withfloats=False, for_frame=False):
        # Build a slow path to call GC write barrier.
        #
        # This builds a helper function called from the fast path of write
        # barriers.  It must save all registers, and optionally all fp
        # registers.  It takes a single argument which is in `r.x10`.  It must
        # keep stack alignment accordingly.

        descr = self.cpu.gc_ll_descr.write_barrier_descr
        if descr is None:
            return

        if not withcards:
            func = descr.get_write_barrier_fn(self.cpu)
        else:
            if descr.jit_wb_cards_set == 0:
                return
            func = descr.get_write_barrier_from_array_fn(self.cpu)
            if func == 0:
                return

        mc = InstrBuilder()

        # Allocate two callee-save scratch registers to handle exception
        # save and restore.
        exc0 = r.x24
        exc1 = r.x25
        fp_align_size = 0
        stack_size = 0
        core_regs_to_be_spilled = []

        if not for_frame:
            self._push_all_regs_to_jitframe(mc, [], withfloats,
                                            callee_only=True)
        else:
            # NOTE: Don't save registers to the jitframe here!  It might
            # override already-saved values that will be restored later.
            #
            # we're possibly called from the slowpath of malloc.  save the
            # caller saved registers assuming GC does not collect here.

            core_regs_to_be_spilled = r.caller_saved_registers + [exc0, exc1]

            core_reg_size = len(core_regs_to_be_spilled) * XLEN
            core_reg_size_aligned = (core_reg_size + FLEN - 1) // FLEN * FLEN
            fp_align_size = core_reg_size_aligned - core_reg_size
            fp_reg_size = len(r.caller_saved_fp_registers) * FLEN
            stack_size = (core_reg_size_aligned + fp_reg_size +
                          ABI_STACK_ALIGN - 1) \
                    // ABI_STACK_ALIGN * ABI_STACK_ALIGN

            mc.ADDI(r.sp.value, r.sp.value, -stack_size)

            # Spill caller-saved registers.
            cur_stack = 0
            for reg in core_regs_to_be_spilled:
                mc.store_int(reg.value, r.sp.value, cur_stack)
                cur_stack += XLEN

            # Spill caller-saved float registers.
            cur_stack += fp_align_size
            for reg in r.caller_saved_fp_registers:
                mc.store_float(reg.value, r.sp.value, cur_stack)
                cur_stack += FLEN

            self._store_and_reset_exception(mc, exc0, exc1)

        func = rffi.cast(lltype.Signed, func)
        mc.load_int_imm(r.ra.value, func)
        mc.JALR(r.ra.value, r.ra.value, 0)

        if not for_frame:
            self._pop_all_regs_from_jitframe(mc, [], withfloats,
                                             callee_only=True)
        else:
            self._restore_exception(mc, exc0, exc1)

            # Restore caller-saved registers.
            cur_stack = 0
            for reg in core_regs_to_be_spilled:
                mc.load_int(reg.value, r.sp.value, cur_stack)
                cur_stack += XLEN

            # Restore caller-saved float registers.
            cur_stack += fp_align_size
            for reg in r.caller_saved_fp_registers:
                mc.load_float(reg.value, r.sp.value, cur_stack)
                cur_stack += FLEN

            mc.ADDI(r.sp.value, r.sp.value, stack_size)

        if withcards:
            # Load and mask the `jit_wb_cards_set_singlebyte` to `x31`, so that
            # the caller of the `wb_slowpath` can emit a simple
            # `BEQZ x31, end_update_card_table`.  This helps us save 2
            # instructions per `COND_CALL_GC_WB_ARRAY`.
            mc.LBU(r.x31.value, r.x10.value, descr.jit_wb_if_flag_byteofs)
            mc.ANDI(r.x31.value, r.x31.value, 0x80)

        mc.RET()
        mc.emit_pending_constants()

        rawstart = mc.materialize(self.cpu, [])
        if for_frame:
            self.wb_slowpath[4] = rawstart
        else:
            self.wb_slowpath[withcards + 2 * withfloats] = rawstart

    def build_frame_realloc_slowpath(self):
        # Build a frame realloc slowpath, which reallocates the frame if the
        # existing frame is smaller than the new size.
        #
        # The slowpath assumes:
        # 1. `r.jfp` holds the old frame address.
        # 2. `r.x31` holds the new frame size.
        # 3. `r.ra` holds the return address

        # Overview: This code should do the following steps:
        #
        # 1. Save all registers to the JITFrame
        # 2. Save exceptions to JITFrame
        # 3. call realloc_frame
        # 4. Set the jfp to point to the new JITFrame
        # 5. Update the JITFrame address on the shadow stack
        # 6. Set the `jf_gcmap` to 0
        # 7. Restore registers
        # 8. Return

        mc = InstrBuilder()

        # Save all registers (except `r.jfp`).
        self._push_all_regs_to_jitframe(mc, [r.jfp], self.cpu.supports_floats)

        # Allocate one callee-saved scratch register for
        # `_store_and_reset_exception`.
        exc_type_reg = r.x25

        # Note: Other backends save the gcmap to `jf_gcmap` here. But in RISCV
        # implementation, we require the caller of this slowpath to set the
        # gcmap so that we don't have to spill another register in the fast
        # path.

        # Set up arguments for `realloc_frame(old_jitframe, new_size)`.
        mc.MV(r.x10.value, r.jfp.value)
        mc.MV(r.x11.value, r.x31.value)

        # Store a possibly present exception.
        self._store_and_reset_exception(mc, None, exc_type_reg,
                                        on_frame=True) # Clobber r.x31 & r.ra

        # Call `realloc_frame(old_jitframe, new_size)`.
        #
        # See also. `rpython/jit/backend/llsupport/llmodel.py` for
        # `realloc_frame`.
        func = rffi.cast(lltype.Signed, self.cpu.realloc_frame)
        mc.load_int_imm(r.ra.value, func)
        mc.JALR(r.ra.value, r.ra.value, 0)

        # Set `r.jfp` to the new JITFrame returned from the previous call.
        mc.MV(r.jfp.value, r.x10.value)

        # Restore a possibly present exception.
        self._restore_exception(mc, None, exc_type_reg)  # Clobber r.ra & r.x31

        # Updates the address at the top of the shadow stack.
        gcrootmap = self.cpu.gc_ll_descr.gcrootmap
        if gcrootmap and gcrootmap.is_shadow_stack:
            scratch_reg = r.x31
            rst = gcrootmap.get_root_stack_top_addr()
            mc.load_int_imm(scratch_reg.value, rst)
            mc.load_int(scratch_reg.value, scratch_reg.value, 0)

            # Update the JITFrame address on the shadow stack.
            mc.store_int(r.jfp.value, scratch_reg.value, -XLEN)

        # Reset the `jf_gcmap`.
        gcmap_ofs = self.cpu.get_ofs_of_frame_field('jf_gcmap')
        mc.store_int(r.x0.value, r.jfp.value, gcmap_ofs)

        # Restore all registers (except `r.jfp`).
        self._pop_all_regs_from_jitframe(mc, [r.jfp], self.cpu.supports_floats)

        # Return
        mc.RET()
        mc.emit_pending_constants()

        rawstart = mc.materialize(self.cpu, [])
        self._frame_realloc_slowpath = rawstart

    def _check_frame_depth(self, mc, gcmap, expected_size):
        # Check if the frame is of enough depth to follow this bridge.
        #
        # If the frame isn't large enough, call `_frame_realloc_slowpath` to
        # enlarge the frame.

        scratch_reg = r.x31
        scratch2_reg = r.ra

        # Load the frame depth from the JITFrame.
        descrs = self.cpu.gc_ll_descr.getframedescrs(self.cpu)
        ofs = self.cpu.unpack_fielddescr(descrs.arraydescr.lendescr)
        mc.load_int(scratch2_reg.value, r.jfp.value, ofs)

        # Load the target for the frame depth.
        if expected_size == -1:
            stack_check_cmp_ofs = mc.get_relative_pos()
            mc.EBREAK()  # Patch Location: LOAD_INT scratch_reg, expected_size
            mc.NOP()
            self._frame_depth_to_patch.append(stack_check_cmp_ofs)
        else:
            mc.load_int_imm(scratch_reg.value, expected_size)

        # Patch Location: `BGE scratch2_reg, scratch_reg, end`
        jg_location = mc.get_relative_pos()
        mc.EBREAK()

        # Store gcmap to frame.
        gcmap_ofs = self.cpu.get_ofs_of_frame_field('jf_gcmap')
        mc.load_int_imm(scratch2_reg.value, rffi.cast(lltype.Signed, gcmap))
        mc.store_int(scratch2_reg.value, r.jfp.value, gcmap_ofs)

        # Call `_frame_realloc_slowpath(x31=new_size)`
        mc.load_int_imm(r.ra.value, self._frame_realloc_slowpath)
        mc.JALR(r.ra.value, r.ra.value, 0)

        # LABEL[end]:

        # Patch the `jg_location` above.
        currpos = mc.get_relative_pos()
        pmc = OverwritingBuilder(mc, jg_location, INST_SIZE)
        pmc.BGE(scratch2_reg.value, scratch_reg.value, currpos - jg_location)

    def check_frame_depth_before_jump(self, target_token):
        if target_token in self.target_tokens_currently_compiling:
            return
        if target_token._riscv_clt is self.current_clt:
            return

        # If we are jumping to another loop or bridge, their frame depth
        # requirement can be larger than what we currently have. Thus, emit
        # `_check_frame_depth` sequence, which enlarges JITFrame if necessary.
        expected_size = target_token._riscv_clt.frame_info.jfi_frame_depth
        gcmap = self._regalloc.get_gcmap()
        self._check_frame_depth(self.mc, gcmap, expected_size)

    def patch_frame_depth_checks(self, frame_depth):
        for ofs in self._frame_depth_to_patch:
            pmc = OverwritingBuilder(self.mc, ofs, INST_SIZE * 2)
            pmc.load_int_imm(r.x31.value, frame_depth)

    def update_frame_depth(self, frame_depth):
        baseofs = self.cpu.get_baseofs_of_frame_field()
        self.current_clt.frame_info.update_frame_depth(baseofs, frame_depth)

    def _store_and_reset_exception(self, mc, exc_val_loc=None,
                                   exc_tp_loc=None, on_frame=False):
        # Move the exception object and type from the addresses provided by
        # `self.cpu.pos_exc_value()` and `self.cpu.pos_excption()` to (1) the
        # specified registers and/or (2) `JITFrame.jf_guard_exc` and then
        # reset the data at the addresses provided by `self.cpu.pos_*()`.

        scratch_reg = r.x31
        assert exc_val_loc is not scratch_reg
        assert exc_tp_loc is not scratch_reg

        # Move the data at `self.cpu.pos_exc_value()` to specified location.
        mc.load_int_imm(scratch_reg.value, self.cpu.pos_exc_value())
        if exc_val_loc is not None:
            assert exc_val_loc.is_core_reg()
            mc.load_int(exc_val_loc.value, scratch_reg.value, 0)
        if on_frame:
            # Store exc_value to the JITFRAME.jf_guard_exc
            scratch2_reg = r.ra  # Clobber r.ra is fine when `on_frame=True`.
            ofs = self.cpu.get_ofs_of_frame_field('jf_guard_exc')
            mc.load_int(scratch2_reg.value, scratch_reg.value, 0)
            mc.store_int(scratch2_reg.value, r.jfp.value, ofs)

        # Reset `self.cpu.pos_exc_value()`.
        mc.store_int(r.x0.value, scratch_reg.value, 0)

        # Move the data at `self.cpu.pos_exception()` to specified location.
        mc.load_int_imm(scratch_reg.value, self.cpu.pos_exception())
        if exc_tp_loc is not None:
            assert exc_tp_loc.is_core_reg()
            mc.load_int(exc_tp_loc.value, scratch_reg.value, 0)

        # Reset `self.cpu.pos_exception()`.
        mc.store_int(r.x0.value, scratch_reg.value, 0)

    def _restore_exception(self, mc, exc_val_loc, exc_tp_loc):
        # Restore `self.cpu.pos_exc_value()` and `self.cpu.pos_exception()`
        # from `exc_val_loc` (or `jf_guard_exc`) and `exc_tp_loc` registers.

        # Allocate scratch registeres.
        scratch_reg = r.x31
        scratch2_reg = r.ra
        assert (exc_val_loc is not scratch_reg and
                exc_val_loc is not scratch2_reg)
        assert (exc_tp_loc is not scratch_reg and
                exc_tp_loc is not scratch2_reg)

        # Restore `pos_exc_value`.
        mc.load_int_imm(scratch_reg.value, self.cpu.pos_exc_value())
        if exc_val_loc is not None:
            mc.store_int(exc_val_loc.value, scratch_reg.value, 0)
        else:
            # Load `exc_value` from JITFRAME and put it in `pos_exc_value`.
            ofs = self.cpu.get_ofs_of_frame_field('jf_guard_exc')
            mc.load_int(scratch2_reg.value, r.jfp.value, ofs)
            mc.store_int(scratch2_reg.value, scratch_reg.value, 0)

            # Reset `jf_guard_exc` in the JITFRAME.
            mc.store_int(r.x0.value, r.jfp.value, ofs)

        # Restore `pos_exception` from `exc_tp_loc`.
        mc.load_int_imm(scratch_reg.value, self.cpu.pos_exception())
        mc.store_int(exc_tp_loc.value, scratch_reg.value, 0)

    def _build_propagate_exception_path(self):
        mc = InstrBuilder()

        # Allocate scratch registers.
        #
        # Note: Use `r.x29` instead of `r.x31` because
        # `_store_and_reset_exception` uses `r.x31` internally.
        scratch_reg = r.x29

        self._store_and_reset_exception(mc, scratch_reg)

        ofs = self.cpu.get_ofs_of_frame_field('jf_guard_exc')
        mc.store_int(scratch_reg.value, r.jfp.value, ofs)

        # Store propagate_exception_descr into frame
        propagate_exception_descr = rffi.cast(
            lltype.Signed,
            cast_instance_to_gcref(self.cpu.propagate_exception_descr))
        ofs = self.cpu.get_ofs_of_frame_field('jf_descr')
        mc.load_int_imm(scratch_reg.value, propagate_exception_descr)
        mc.store_int(scratch_reg.value, r.jfp.value, ofs)

        self._call_footer(mc)
        mc.emit_pending_constants()

        rawstart = mc.materialize(self.cpu, [])
        self.propagate_exception_path = rawstart

    def _build_cond_call_slowpath(self, supports_floats, callee_only):
        """ This builds a general call slowpath, for whatever call happens to
        come.

        The address of the callee function comes in r.x30.
        The returning value is stored in r.x30.
        """
        mc = InstrBuilder()

        # Spill registers to JITFRAME
        #
        # Ignore jfp for _reload_frame_if_necessary, x30 for return, x31 for
        # scratch.
        ignore_regs_for_push_pop = [r.jfp, r.x30, r.x31]
        self._push_all_regs_to_jitframe(mc, ignore_regs_for_push_pop,
                                        supports_floats,
                                        callee_only)  # Spills r.ra

        # Branch to the callee function.
        mc.JALR(r.ra.value, r.x30.value, 0)

        # Move return value to r.x30.
        mc.MV(r.x30.value, r.x10.value)

        # Restore registers from JITFRAME
        tmplocs = [r.x29]  # Use callee-saved register as scratch regs
        self._reload_frame_if_necessary(mc, tmplocs)
        self._pop_all_regs_from_jitframe(mc, ignore_regs_for_push_pop,
                                         supports_floats,
                                         callee_only)  # Restores r.ra
        mc.RET()
        mc.emit_pending_constants()
        return mc.materialize(self.cpu, [])

    def _reload_frame_if_necessary(self, mc, tmplocs):
        gcrootmap = self.cpu.gc_ll_descr.gcrootmap
        if gcrootmap and gcrootmap.is_shadow_stack:
            stack_top_ptr_addr = gcrootmap.get_root_stack_top_addr()
            mc.load_int_imm(r.jfp.value, stack_top_ptr_addr)
            mc.load_int(r.jfp.value, r.jfp.value, 0)
            mc.load_int(r.jfp.value, r.jfp.value, -XLEN)
        wbdescr = self.cpu.gc_ll_descr.write_barrier_descr
        if gcrootmap and wbdescr:
            # Frame never uses card marking, so we enforce this is not an
            # array.
            self._write_barrier_fastpath(mc, wbdescr, [r.jfp], tmplocs,
                                         array=False, is_frame=True)

    def _build_malloc_slowpath(self, kind):
        # malloc_slowpath for various kinds (fixed, str, unicode, var):
        #
        # x10, x11 = malloc_slowpath_fixed(x10=nursery_free_adr,
        #                                  x11=(nursery_free_adr + size)
        #                                  x31=gcmap)
        #
        # x10, x11 = malloc_slowpath_str/unicode(x10=length_of_string,
        #                                        x31=gcmap)
        #
        # x10, x11 = malloc_slowpath_var(x10=itemsize,
        #                                x11=tid,
        #                                x12=length_of_array,
        #                                x31=gcmap)
        #
        # Returns:
        # x10 = new_object_adr
        # X11 = new_nursery_free_adr

        assert kind in ['fixed', 'str', 'unicode', 'var']
        mc = InstrBuilder()

        # Push registers to JITFrame.

        # Ignore fp for _reload_frame_if_necessary, x10-12 for args, x31 for
        # scratch.
        if kind == 'var':
            ignore_regs_for_push_pop = [r.jfp, r.x10, r.x11, r.x12, r.x31]
        else:
            ignore_regs_for_push_pop = [r.jfp, r.x10, r.x11, r.x31]

        self._push_all_regs_to_jitframe(mc, ignore_regs_for_push_pop,
                                        self.cpu.supports_floats)

        # Select the callee function according to the `kind`.
        if kind == 'fixed':
            addr = self.cpu.gc_ll_descr.get_malloc_slowpath_addr()
        elif kind == 'str':
            addr = self.cpu.gc_ll_descr.get_malloc_fn_addr('malloc_str')
        elif kind == 'unicode':
            addr = self.cpu.gc_ll_descr.get_malloc_fn_addr('malloc_unicode')
        else:
            addr = self.cpu.gc_ll_descr.get_malloc_slowpath_array_addr()

        # Setup the arguments.
        if kind == 'fixed':
            # malloc_slowpath_addr(x10=size)
            # malloc_slowpath_addr(x10=size, x11=jfp)

            # At this point we know that the values we need to compute the size
            # are stored in `r.x10` and `r.x11`.
            mc.SUB(r.x10.value, r.x11.value, r.x10.value)
            if hasattr(self.cpu.gc_ll_descr, 'passes_frame'):
                mc.MV(r.x11.value, r.jfp.value)
        elif kind == 'str' or kind == 'unicode':
            # malloc_str(x10=len), malloc_unicode(x10=len)
            pass
        else:  # var
            # malloc_slowpath_array_addr(x10=itemsize, x11=tid, x12=len)
            pass

        # Store `gcmap` to `jf_gcmap`.
        jf_gcmap_ofs = self.cpu.get_ofs_of_frame_field('jf_gcmap')
        mc.store_int(r.x31.value, r.jfp.value, jf_gcmap_ofs)

        # Call the callee function
        mc.load_int_imm(r.ra.value, rffi.cast(lltype.Signed, addr))
        mc.JALR(r.ra.value, r.ra.value, 0)

        # Patch Loation: BNEZ x10, succeeded
        branch_inst_pos = mc.get_relative_pos()
        mc.EBREAK()

        # If the slowpath malloc failed, we raise a MemoryError that always
        # interrupts the current loop, as a "good enough" approximation.
        mc.load_int_imm(r.ra.value, self.propagate_exception_path)
        mc.JR(r.ra.value)

        # LABEL[succeeded]:
        currpos = mc.get_relative_pos()
        pmc = OverwritingBuilder(mc, branch_inst_pos, INST_SIZE)
        pmc.BNEZ(r.x10.value, currpos - branch_inst_pos)

        # Allocate another caller-save as a scratch register.
        #
        # This must not be `r.ra` nor `r.x31` because `_write_barrier_fastpath`
        # has used them. This can be any other register saved by
        # `_push_all_regs_to_jitframe`.
        scratch2_reg = r.x30

        # Reload the frame.
        self._reload_frame_if_necessary(mc, tmplocs=[scratch2_reg])

        # Pop registers from JITFrame.
        self._pop_all_regs_from_jitframe(mc, ignore_regs_for_push_pop,
                                         self.cpu.supports_floats)

        # Load the nursery_free_adr back to r.x11 because the fast path will
        # store the value in `r.x11` to `&nursery_free_adr`.
        nursery_free_adr = self.cpu.gc_ll_descr.get_nursery_free_addr()
        mc.load_int_imm(r.x11.value, nursery_free_adr)
        mc.load_int(r.x11.value, r.x11.value, 0)

        # Clear the `jf_gcmap`.
        mc.store_int(r.x0.value, r.jfp.value, jf_gcmap_ofs)

        mc.RET()
        mc.emit_pending_constants()

        rawstart = mc.materialize(self.cpu, [])
        return rawstart

    def malloc_cond(self, nursery_free_adr, nursery_top_adr, size, gcmap):
        assert size & (XLEN - 1) == 0

        # Load nursery_free_adr
        self.mc.load_int_imm(r.x10.value, nursery_free_adr)
        self.mc.load_int(r.x10.value, r.x10.value, 0)

        # Add the size to be allocated
        if check_imm_arg(size):
            self.mc.ADDI(r.x11.value, r.x10.value, size)
        else:
            self.mc.load_int_imm(r.x11.value, size)
            self.mc.ADD(r.x11.value, r.x10.value, r.x11.value)

        # Load nursery_top_adr
        scratch_reg = r.x31
        self.mc.load_int_imm(scratch_reg.value, nursery_top_adr)
        self.mc.load_int(scratch_reg.value, scratch_reg.value, 0)

        # Patch Location: BGEU scratch_reg, x11, end
        branch_inst_pos = self.mc.get_relative_pos()
        self.mc.EBREAK()

        # x10, x11 = malloc_slowpath(x10=nursery_free_addr,
        #                            x11=(nursery_free_addr + size),
        #                            x31=gcmap)
        #
        # Returns:
        #
        # x10: new object address
        # X11: new nursery_free_adr

        self.mc.load_int_imm(r.x31.value, rffi.cast(lltype.Signed, gcmap))

        self.mc.load_int_imm(r.ra.value, self.malloc_slowpath)
        self.mc.JALR(r.ra.value, r.ra.value, 0)

        # LABEL[end]:
        currpos = self.mc.get_relative_pos()
        pmc = OverwritingBuilder(self.mc, branch_inst_pos, INST_SIZE)
        pmc.BGEU(scratch_reg.value, r.x11.value, currpos - branch_inst_pos)

        # Update `nursery_free_adr` after allocation.
        self.mc.load_int_imm(scratch_reg.value, nursery_free_adr)
        self.mc.store_int(r.x11.value, scratch_reg.value, 0)

    def malloc_cond_varsize_frame(self, nursery_free_adr, nursery_top_adr,
                                  size_loc, gcmap):
        assert size_loc.is_core_reg()
        assert size_loc is not r.x10 and size_loc is not r.x11

        # Load nursery_free_adr
        self.mc.load_int_imm(r.x10.value, nursery_free_adr)
        self.mc.load_int(r.x10.value, r.x10.value, 0)

        # Add the size to be allocated
        self.mc.ADD(r.x11.value, r.x10.value, size_loc.value)

        # Load nursery_top_adr
        scratch_reg = r.x31
        self.mc.load_int_imm(scratch_reg.value, nursery_top_adr)
        self.mc.load_int(scratch_reg.value, scratch_reg.value, 0)

        # Patch Location: BGEU scratch_reg, x11, end
        branch_inst_pos = self.mc.get_relative_pos()
        self.mc.EBREAK()

        # x10, x11 = malloc_slowpath(x10=nursery_free_addr,
        #                            x11=(nursery_free_addr + size),
        #                            x31=gcmap)

        self.mc.load_int_imm(r.x31.value, rffi.cast(lltype.Signed, gcmap))

        self.mc.load_int_imm(r.ra.value, self.malloc_slowpath)
        self.mc.JALR(r.ra.value, r.ra.value, 0)

        # LABEL[end]:
        currpos = self.mc.get_relative_pos()
        pmc = OverwritingBuilder(self.mc, branch_inst_pos, INST_SIZE)
        pmc.BGEU(scratch_reg.value, r.x11.value, currpos - branch_inst_pos)

        self.mc.load_int_imm(scratch_reg.value, nursery_free_adr)
        self.mc.store_int(r.x11.value, scratch_reg.value, 0)

    def malloc_cond_varsize(self, kind, nursery_free_adr, nursery_top_adr,
                            length_loc, itemsize, max_length, gcmap,
                            arraydescr):
        assert isinstance(arraydescr, ArrayDescr)

        scratch_reg = r.x31

        self.mc.load_int_imm(scratch_reg.value, max_length)

        # Patch Location: BLT scratch_reg, length, call_slowpath
        jmp_adr0 = self.mc.get_relative_pos()
        self.mc.EBREAK()

        # Load nursery_free_adr to r.x10
        self.mc.load_int_imm(r.x10.value, nursery_free_adr)
        self.mc.load_int(r.x10.value, r.x10.value, 0)

        # Calculate total size (header_size + itemsize * len) to be allocated
        self.mc.load_int_imm(scratch_reg.value, itemsize)
        # x11 = length * itemsize
        self.mc.MUL(r.x11.value, length_loc.value, scratch_reg.value)

        assert arraydescr.basesize >= self.gc_minimal_size_in_nursery
        constsize = arraydescr.basesize + self.gc_size_of_header
        force_realignment = (itemsize % XLEN) != 0
        if force_realignment:
            constsize += XLEN - 1
        # x11 = x11 + constsize
        if check_imm_arg(constsize):
            self.mc.ADDI(r.x11.value, r.x11.value, constsize)
        else:
            self.mc.load_int_imm(scratch_reg.value, constsize)
            self.mc.ADD(r.x11.value, r.x11.value, scratch_reg.value)

        # Calculate new nursery_free_adr
        self.mc.ADD(r.x11.value, r.x11.value, r.x10.value)
        if force_realignment:
            self.mc.ANDI(r.x11.value, r.x11.value, -XLEN)

        # Load nursery_top_adr
        self.mc.load_int_imm(scratch_reg.value, nursery_top_adr)
        self.mc.load_int(scratch_reg.value, scratch_reg.value, 0)

        # Patch Location: BGEU scratch_reg, x11, finish_fast_alloc
        jmp_adr1 = self.mc.get_relative_pos()
        self.mc.EBREAK()

        # LABEL[call_slowpath]:
        currpos = self.mc.get_relative_pos()
        pmc = OverwritingBuilder(self.mc, jmp_adr0, INST_SIZE)
        pmc.BLT(scratch_reg.value, length_loc.value, currpos - jmp_adr0)

        # Setup the arguments to slowpaths (see also. _build_malloc_slowpath)
        if kind == rewrite.FLAG_ARRAY:
            self.mc.load_int_imm(r.x10.value, itemsize)
            self.mc.load_int_imm(r.x11.value, arraydescr.tid)
            self.regalloc_mov(length_loc, r.x12)
            addr = self.malloc_slowpath_varsize
        else:
            if kind == rewrite.FLAG_STR:
                addr = self.malloc_slowpath_str
            else:
                assert kind == rewrite.FLAG_UNICODE
                addr = self.malloc_slowpath_unicode
            self.regalloc_mov(length_loc, r.x10)

        # Load the gcmap to r.x31
        self.mc.load_int_imm(r.x31.value, rffi.cast(lltype.Signed, gcmap))

        # Call the callee
        self.mc.load_int_imm(r.ra.value, addr)
        self.mc.JALR(r.ra.value, r.ra.value, 0)

        # Patch Location: J done
        jmp_location = self.mc.get_relative_pos()
        self.mc.EBREAK()

        # LABEL[finish_fast_alloc]:
        currpos = self.mc.get_relative_pos()
        pmc = OverwritingBuilder(self.mc, jmp_adr1, INST_SIZE)
        pmc.BGEU(scratch_reg.value, r.x11.value, currpos - jmp_adr1)

        # Write down the tid.
        self.mc.load_int_imm(scratch_reg.value, arraydescr.tid)
        self.mc.store_int(scratch_reg.value, r.x10.value, 0)

        # Write the new `nursery_free_adr`.
        self.mc.load_int_imm(scratch_reg.value, nursery_free_adr)
        self.mc.store_int(r.x11.value, scratch_reg.value, 0)

        # LABEL[done]:
        currpos = self.mc.get_relative_pos()
        pmc = OverwritingBuilder(self.mc, jmp_location, INST_SIZE)
        pmc.J(currpos - jmp_location)

    def _build_stack_check_slowpath(self):
        _, _, slowpathaddr = self.cpu.insert_stack_check()
        if slowpathaddr == 0 or not self.cpu.propagate_exception_descr:
            return  # No stack check (for tests, or non-translated)

        # Make a "function" that is called immediately at the start of
        # an assembler function.  In particular, the stack looks like:
        #
        #    | saved argument regs |
        #    | retaddr             |  <-- sp
        #    +---------------------+
        #
        mc = InstrBuilder()

        # Save argument registers and return address
        stack_size = (((len(r.argument_regs) + 1) * XLEN + ABI_STACK_ALIGN - 1)
                      // ABI_STACK_ALIGN * ABI_STACK_ALIGN)

        mc.ADDI(r.sp.value, r.sp.value, -stack_size)
        mc.store_int(r.ra.value, r.sp.value, 0)
        for i in range(len(r.argument_regs)):
            mc.store_int(r.argument_regs[i].value, r.sp.value, (i + 1) * XLEN)

        # Pass current stack pointer as argument to the call
        mc.MV(r.x10.value, r.sp.value)
        mc.load_int_imm(r.ra.value, slowpathaddr)
        mc.JALR(r.ra.value, r.ra.value, 0)

        # Check for an exception
        mc.load_int_imm(r.x10.value, self.cpu.pos_exception())
        mc.load_int(r.x10.value, r.x10.value, 0)

        # Patch Location: BNEZ r.x10, propagate_exc
        jmp = mc.get_relative_pos()
        mc.EBREAK()

        # Restore registers and return
        for i in range(len(r.argument_regs)):
            mc.load_int(r.argument_regs[i].value, r.sp.value, (i + 1) * XLEN)
        mc.load_int(r.ra.value, r.sp.value, 0)
        mc.ADDI(r.sp.value, r.sp.value, stack_size)
        mc.RET()

        # LABEL[propagate_exc]:
        pmc = OverwritingBuilder(mc, jmp, INST_SIZE)
        pmc.BNEZ(r.x10.value, mc.get_relative_pos() - jmp)

        mc.ADDI(r.sp.value, r.sp.value, stack_size)
        mc.load_int_imm(r.ra.value, self.propagate_exception_path)
        mc.JR(r.ra.value)
        mc.emit_pending_constants()

        rawstart = mc.materialize(self.cpu, [])
        self.stack_check_slowpath = rawstart

    def load_imm(self, loc, imm):
        """Load an immediate value into a register"""
        if loc.is_core_reg():
            assert imm.is_imm()
            self.mc.load_int_imm(loc.value, imm.value)
        else:
            assert loc.is_fp_reg() and imm.is_imm_float()
            self.mc.load_float_imm(loc.value, imm.value)

    def regalloc_mov(self, prev_loc, loc):
        """Moves a value from a previous location to some other location"""
        if prev_loc.is_imm():
            return self._mov_imm_to_loc(prev_loc, loc)
        elif prev_loc.is_stack():
            self._mov_stack_to_loc(prev_loc, loc)
        elif prev_loc.is_core_reg():
            self._mov_reg_to_loc(prev_loc, loc)
        elif prev_loc.is_fp_reg():
            self._mov_fp_reg_to_loc(prev_loc, loc)
        elif prev_loc.is_imm_float():
            self._mov_imm_float_to_loc(prev_loc, loc)
        else:
            assert 0, 'unsupported case'
    mov_loc_loc = regalloc_mov

    def _mov_imm_to_loc(self, prev_loc, loc):
        if loc.is_core_reg():
            self.mc.load_int_imm(loc.value, prev_loc.value)
        else:
            assert 0, 'unsupported case'

    def _mov_stack_to_loc(self, prev_loc, loc):
        offset = prev_loc.value
        if loc.is_core_reg():
            self.mc.load_int_from_base_plus_offset(loc.value, r.jfp.value,
                                                   offset)
        elif loc.is_fp_reg():
            self.mc.load_float_from_base_plus_offset(loc.value, r.jfp.value,
                                                     offset, tmp=r.x31.value)
        else:
            assert 0, 'unsupported case'

    def _mov_reg_to_loc(self, prev_loc, loc):
        if loc.is_core_reg():
            self.mc.MV(loc.value, prev_loc.value)
        elif loc.is_stack():
            # Use `r.shadow_old` as `scratch_reg`.  We can't use `r.x31`
            # because `prev_loc` can be `r.x31` (see also.
            # `regalloc_prepare_move`).  We can't use `r.ra` because `r.ra` is
            # allocated for the callee function address in `callbuiler.py`
            # and its lifetime overlaps with `remap_frame_layout`.
            scratch_reg = r.shadow_old
            self.mc.store_int_to_base_plus_offset(prev_loc.value, r.jfp.value,
                                                  loc.value,
                                                  tmp=scratch_reg.value)
        else:
            assert 0, 'unsupported case'

    def _mov_fp_reg_to_loc(self, prev_loc, loc):
        if loc.is_fp_reg():
            self.mc.FMV_D(loc.value, prev_loc.value)
        elif loc.is_core_reg():
            assert XLEN == 8 and FLEN == 8
            self.mc.FMV_X_D(loc.value, prev_loc.value)
        elif loc.is_stack():
            self.mc.store_float_to_base_plus_offset(prev_loc.value,
                                                    r.jfp.value, loc.value,
                                                    tmp=r.x31.value)
        else:
            assert 0, 'unsupported case'

    def _mov_imm_float_to_loc(self, prev_loc, loc):
        if loc.is_fp_reg():
            self.mc.load_float_imm(loc.value, prev_loc.value)
        elif loc.is_stack():
            self.mc.load_float_imm(r.f31.value, prev_loc.value)
            self.mc.store_float_to_base_plus_offset(r.f31.value, r.jfp.value,
                                                    loc.value, tmp=r.x31.value)
        else:
            assert 0, 'unsupported case'

    def mov_loc_to_raw_stack(self, loc, sp_offset):
        # Move a value to sp[sp_offset], which is usually for foreign function
        # calls.
        if loc.is_core_reg():
            self.mc.store_int(loc.value, r.sp.value, sp_offset)
        elif loc.is_stack():
            # Move a value from JITFRAME stack to raw stack.
            scratch_reg = r.x31
            self.mc.load_int_from_base_plus_offset(scratch_reg.value,
                                                   r.jfp.value, loc.value)
            self.mc.store_int(scratch_reg.value, r.sp.value, sp_offset)
        elif loc.is_fp_reg():
            self.mc.store_float(loc.value, r.sp.value, sp_offset)
        elif loc.is_imm():
            scratch_reg = r.x31
            self.mc.load_int_imm(scratch_reg.value, loc.value)
            self.mc.store_int(scratch_reg.value, r.sp.value, sp_offset)
        else:
            assert 0, 'unsupported case'

    def regalloc_push(self, loc, already_pushed):
        """Push the value stored in `loc` to the stack top.

        Side effect: r.x31 or r.f31 may be overwritten."""

        offset = SCRATCH_STACK_SLOT_SIZE * (~already_pushed)

        if loc.type == FLOAT:
            if not loc.is_fp_reg():
                self.regalloc_mov(loc, r.f31)
                loc = r.f31
            self.mc.store_float(loc.value, r.sp.value, offset)
        else:
            if not loc.is_core_reg():
                self.regalloc_mov(loc, r.x31)
                loc = r.x31
            self.mc.store_int(loc.value, r.sp.value, offset)

    def regalloc_pop(self, loc, already_pushed):
        """Pop the value from the top of the stack to `loc`.

        Side effect: r.x31 or r.f31 may be overwritten."""

        offset = SCRATCH_STACK_SLOT_SIZE * (~already_pushed)

        if loc.type == FLOAT:
            if loc.is_fp_reg():
                self.mc.load_float(loc.value, r.sp.value, offset)
            else:
                self.mc.load_float(r.f31.value, r.sp.value, offset)
                self.regalloc_mov(r.f31, loc)
        else:
            if loc.is_core_reg():
                self.mc.load_int(loc.value, r.sp.value, offset)
            else:
                self.mc.load_int(r.x31.value, r.sp.value, offset)
                self.regalloc_mov(r.x31, loc)

    def regalloc_prepare_move(self, src, dst, tmp):
        """Move `src` to `tmp` and return `tmp` if `src`-to-`dst` is a
        stack-to-stack or imm-to-stack move."""
        if dst.is_stack() and (src.is_stack() or src.is_imm()):
            self.regalloc_mov(src, tmp)
            return tmp
        return src

    def imm(self, value):
        return ImmLocation(value)

    def new_stack_loc(self, i, tp):
        # Create a StackLocation at `i` of type `tp`.
        #
        # Note: This function is called by rebuild_faillocs_from_descr()

        base_ofs = self.cpu.get_baseofs_of_frame_field()
        return StackLocation(i, get_fp_offset(base_ofs, i), tp)



"""
Semi-public interface to gather and print a raw traceback, e.g.
from the faulthandler module.
"""

from rpython.rlib.rvmprof import cintf, rvmprof
from rpython.rlib.objectmodel import specialize
from rpython.rtyper.lltypesystem import lltype, llmemory, rffi
from rpython.rlib.rarithmetic import widen


def traceback(estimate_number_of_entries):
    """Build and return a vmprof-like traceback, as a pair (array_p,
    array_length).  The caller must free array_p.  Not for signal handlers:
    for these, call vmprof_get_traceback() from C code.
    """
    if not cintf.IS_SUPPORTED:
        return (None, 0)
    _cintf = rvmprof._get_vmprof().cintf
    size = estimate_number_of_entries * 2 + 4
    stack = cintf.get_rvmprof_stack()
    array_p = lltype.malloc(rffi.SIGNEDP.TO, size, flavor='raw')
    NULL = llmemory.NULL
    voidpp = rffi.cast(rffi.VOIDPP, array_p)
    array_length = _cintf.vmprof_get_traceback(stack, NULL, voidpp, size)
    return (array_p, array_length)


LOC_INTERPRETED    = 0
LOC_JITTED         = 1
LOC_JITTED_INLINED = 2


@specialize.arg(0, 1)
def _traceback_one(CodeClass, callback, arg, code_id, loc):
    found_code = None
    code_id = widen(code_id)
    if code_id != 0:
        all_code_wrefs = CodeClass._vmprof_weak_list.get_all_handles()
        i = len(all_code_wrefs) - 1
        while i >= 0:
            code = all_code_wrefs[i]()
            if code is not None and code._vmprof_unique_id == code_id:
                found_code = code
                break
            i -= 1
    callback(found_code, loc, arg)

@specialize.arg(0, 1)
def walk_traceback(CodeClass, callback, arg, array_p, array_length):
    """Invoke 'callback(code_obj, loc, arg)' for every traceback entry.
    'code_obj' may be None if it can't be determined.  'loc' is one
    of the LOC_xxx constants.
    """
    if not cintf.IS_SUPPORTED:
        return
    i = 0
    array_length = widen(array_length)
    array_p = rffi.cast(rffi.INTPTR_TP, array_p)
    while i < array_length - 1:
        tag = array_p[i]
        tagged_value = array_p[i + 1]
        if tag == rvmprof.VMPROF_CODE_TAG:
            loc = LOC_INTERPRETED
            _traceback_one(CodeClass, callback, arg, tagged_value, loc)
        elif tag == rvmprof.VMPROF_JITTED_TAG:
            if i + 2 >= array_length:  # skip last entry, can't determine if
                break                  # it's LOC_JITTED_INLINED or LOC_JITTED
            if widen(array_p[i + 2]) == rvmprof.VMPROF_JITTED_TAG:
                loc = LOC_JITTED_INLINED
            else:
                loc = LOC_JITTED
            _traceback_one(CodeClass, callback, arg, tagged_value, loc)
        i += 2



from rpython.jit.codewriter.effectinfo import EffectInfo
from rpython.jit.codewriter import longlong
from rpython.jit.metainterp import compile
from rpython.jit.metainterp.history import (
    Const, ConstInt, make_hashable_int, ConstFloat, CONST_NULL,
    DONT_CHANGE)
from rpython.jit.metainterp.optimize import InvalidLoop
from rpython.jit.metainterp.optimizeopt.optimizer import (
    Optimization, OptimizationResult, REMOVED, CONST_0, CONST_1)
from rpython.jit.metainterp.optimizeopt.info import (
    INFO_NONNULL, INFO_NULL, getptrinfo, ArrayPtrInfo)
from rpython.jit.metainterp.optimizeopt.util import (
    _findall, make_dispatcher_method, have_dispatcher_method, get_box_replacement)
from rpython.jit.metainterp.resoperation import (
    rop, ResOperation, opclasses, OpHelpers)
from rpython.rlib.rarithmetic import highest_bit, LONG_BIT
import math


class CallLoopinvariantOptimizationResult(OptimizationResult):
    def __init__(self, opt, op, old_op):
        OptimizationResult.__init__(self, opt, op)
        self.old_op = old_op

    def callback(self):
        self._callback(self.op, self.old_op)

    def _callback(self, op, old_op):
        key = make_hashable_int(op.getarg(0).getint())
        self.opt.loop_invariant_producer[key] = self.opt.optimizer.getlastop()
        self.opt.loop_invariant_results[key] = old_op


class OptRewrite(Optimization):
    """Rewrite operations into equivalent, cheaper operations.
       This includes already executed operations and constants.
    """
    def __init__(self):
        self.loop_invariant_results = {}
        self.loop_invariant_producer = {}

    def setup(self):
        self.optimizer.optrewrite = self

    def produce_potential_short_preamble_ops(self, sb):
        for op in self.loop_invariant_producer.values():
            sb.add_loopinvariant_op(op)

    def propagate_forward(self, op):
        if opclasses[op.opnum].boolinverse != -1 or opclasses[op.opnum].boolreflex != -1:
            if self.find_rewritable_bool(op):
                return

        return dispatch_opt(self, op)

    def try_boolinvers(self, op, targs):
        oldop = self.get_pure_result(targs)
        if oldop is not None:
            b = self.getintbound(oldop)
            if b.known_eq_const(1):
                self.make_constant(op, CONST_0)
                return True
            elif b.known_eq_const(0):
                self.make_constant(op, CONST_1)
                return True
        return False

    def find_rewritable_bool(self, op):
        oldopnum = op.boolinverse
        arg0 = op.getarg(0)
        arg1 = op.getarg(1)
        if oldopnum != -1:
            top = ResOperation(oldopnum, [arg0, arg1])
            if self.try_boolinvers(op, top):
                return True

        oldopnum = op.boolreflex  # FIXME: add INT_ADD, INT_MUL
        if oldopnum != -1:
            top = ResOperation(oldopnum, [arg1, arg0])
            oldop = self.get_pure_result(top)
            if oldop is not None:
                self.optimizer.make_equal_to(op, oldop)
                return True

        if op.boolreflex == -1:
            return False
        oldopnum = opclasses[op.boolreflex].boolinverse
        if oldopnum != -1:
            top = ResOperation(oldopnum, [arg1, arg0])
            if self.try_boolinvers(op, top):
                return True

        return False

    def _optimize_CALL_INT_UDIV(self, op):
        b2 = self.getintbound(op.getarg(2))
        if b2.is_constant() and b2.get_constant_int() == 1:
            self.make_equal_to(op, op.getarg(1))
            self.last_emitted_operation = REMOVED
            return True
        return False

    def optimize_FLOAT_MUL(self, op):
        arg1 = op.getarg(0)
        arg2 = op.getarg(1)

        # Constant fold f0 * 1.0 and turn f0 * -1.0 into a FLOAT_NEG, these
        # work in all cases, including NaN and inf
        for lhs, rhs in [(arg1, arg2), (arg2, arg1)]:
            v1 = get_box_replacement(lhs)
            v2 = get_box_replacement(rhs)

            if v1.is_constant():
                if v1.getfloat() == 1.0:
                    self.make_equal_to(op, v2)
                    return
                elif v1.getfloat() == -1.0:
                    newop = self.replace_op_with(op, rop.FLOAT_NEG, args=[rhs])
                    return self.emit(newop)
        return self.emit(op)

    def postprocess_FLOAT_MUL(self, op):
        self.optimizer.pure_from_args2(rop.FLOAT_MUL,
                                       op.getarg(1), op.getarg(0), op)

    def optimize_FLOAT_TRUEDIV(self, op):
        arg1 = op.getarg(0)
        arg2 = op.getarg(1)
        v2 = get_box_replacement(arg2)

        # replace "x / const" by "x * (1/const)" if possible
        newop = op
        if v2.is_constant():
            divisor = v2.getfloat()
            fraction = math.frexp(divisor)[0]
            # This optimization is valid for powers of two
            # but not for zeroes, some denormals and NaN:
            if fraction == 0.5 or fraction == -0.5:
                reciprocal = 1.0 / divisor
                rfraction = math.frexp(reciprocal)[0]
                if rfraction == 0.5 or rfraction == -0.5:
                    c = ConstFloat(longlong.getfloatstorage(reciprocal))
                    newop = self.replace_op_with(op, rop.FLOAT_MUL,
                                                 args=[arg1, c])
        return self.emit(newop)

    def optimize_FLOAT_NEG(self, op):
        v = get_box_replacement(op.getarg(0))
        arg_op = self.optimizer.as_operation(v, rop.FLOAT_NEG)
        if arg_op is not None:
            self.make_equal_to(op, arg_op.getarg(0))
        else:
            return self.emit(op)

    def optimize_FLOAT_ABS(self, op):
        v = get_box_replacement(op.getarg(0))
        arg_op = self.optimizer.as_operation(v, rop.FLOAT_ABS)
        if arg_op is not None:
            self.make_equal_to(op, v)
        else:
            return self.emit(op)

    def optimize_guard(self, op, constbox):
        box = op.getarg(0)
        if box.type == 'i':
            intbound = self.getintbound(box)
            if intbound.is_constant():
                if not intbound.get_constant_int() == constbox.getint():
                    r = self.optimizer.metainterp_sd.logger_ops.repr_of_resop(
                        op)
                    raise InvalidLoop('A GUARD_{VALUE,TRUE,FALSE} (%s) '
                                      'was proven to always fail' % r)
                return
        elif box.type == 'r':
            box = get_box_replacement(box)
            if box.is_constant():
                if not box.same_constant(constbox):
                    r = self.optimizer.metainterp_sd.logger_ops.repr_of_resop(
                        op)
                    raise InvalidLoop('A GUARD_VALUE (%s) '
                                      'was proven to always fail' % r)
                return

        return self.emit(op)

    def optimize_GUARD_ISNULL(self, op):
        info = getptrinfo(op.getarg(0))
        if info is not None:
            if info.is_null():
                return
            elif info.is_nonnull():
                r = self.optimizer.metainterp_sd.logger_ops.repr_of_resop(op)
                raise InvalidLoop('A GUARD_ISNULL (%s) was proven to always '
                                  'fail' % r)
        return self.emit(op)

    def postprocess_GUARD_ISNULL(self, op):
        self.make_constant(op.getarg(0), CONST_NULL)

    def optimize_GUARD_IS_OBJECT(self, op):
        info = getptrinfo(op.getarg(0))
        if info and info.is_constant():
            if info.is_null():
                raise InvalidLoop("A GUARD_IS_OBJECT(NULL) found")
            c = get_box_replacement(op.getarg(0))
            if self.optimizer.cpu.check_is_object(c.getref_base()):
                return
            raise InvalidLoop("A GUARD_IS_OBJECT(not-an-object) found")
        if info is not None:
            if info.is_about_object():
                return
            if info.is_precise():
                raise InvalidLoop()
        return self.emit(op)

    def optimize_GUARD_GC_TYPE(self, op):
        info = getptrinfo(op.getarg(0))
        if info and info.is_constant():
            c = get_box_replacement(op.getarg(0))
            tid = self.optimizer.cpu.get_actual_typeid(c.getref_base())
            if tid != op.getarg(1).getint():
                raise InvalidLoop("wrong GC type ID found on a constant")
            return
        if info is not None and info.get_descr() is not None:
            if info.get_descr().get_type_id() != op.getarg(1).getint():
                raise InvalidLoop("wrong GC types passed around!")
            return
        if info is not None and info.is_about_object():
            known_class = info.get_known_class(self.optimizer.cpu)
            if known_class:
                try:
                    tid = self.optimizer.metainterp_sd._try_find_typeid_for_vtable(known_class)
                except KeyError:
                    pass
                else:
                    if tid != op.getarg(1).getint():
                        raise InvalidLoop("guard_class and guard_gc_type contradict each other")
        return self.emit(op)

    def optimize_GUARD_SUBCLASS(self, op):
        info = getptrinfo(op.getarg(0))
        optimizer = self.optimizer
        # must raise 'InvalidLoop' in all cases where 'info' shows the
        # class cannot possibly match (see test_issue2926)
        if info and info.is_constant():
            c = get_box_replacement(op.getarg(0))
            vtable = optimizer.cpu.cls_of_box(c).getint()
            if optimizer._check_subclass(vtable, op.getarg(1).getint()):
                return
            raise InvalidLoop("GUARD_SUBCLASS(const) proven to always fail")
        if info is not None and info.is_about_object():
            known_class = info.get_known_class(optimizer.cpu)
            if known_class:
                # Class of 'info' is exactly 'known_class'.
                # We know statically if the 'guard_subclass' will pass or fail.
                if optimizer._check_subclass(known_class.getint(),
                                             op.getarg(1).getint()):
                    return
                else:
                    raise InvalidLoop(
                        "GUARD_SUBCLASS(known_class) proven to always fail")
            elif info.get_descr() is not None:
                # Class of 'info' is either get_descr() or a subclass of it.
                # We're keeping the 'guard_subclass' at runtime only in the
                # case where get_descr() is some strict parent class of
                # the argument to 'guard_subclass'.
                info_base_descr = info.get_descr().get_vtable()
                if optimizer._check_subclass(info_base_descr,
                                             op.getarg(1).getint()):
                    return    # guard_subclass always passing
                elif optimizer._check_subclass(op.getarg(1).getint(),
                                               info_base_descr):
                    pass      # don't know, must keep the 'guard_subclass'
                else:
                    raise InvalidLoop(
                        "GUARD_SUBCLASS(base_class) proven to always fail")
        return self.emit(op)

    def optimize_GUARD_NONNULL(self, op):
        opinfo = getptrinfo(op.getarg(0))
        if opinfo is not None:
            if opinfo.is_nonnull():
                return
            elif opinfo.is_null():
                r = self.optimizer.metainterp_sd.logger_ops.repr_of_resop(op)
                raise InvalidLoop('A GUARD_NONNULL (%s) was proven to always '
                                  'fail' % r)
        return self.emit(op)

    def postprocess_GUARD_NONNULL(self, op):
        self.make_nonnull(op.getarg(0))
        getptrinfo(op.getarg(0)).mark_last_guard(self.optimizer)

    def optimize_GUARD_VALUE(self, op):
        arg0 = op.getarg(0)
        if arg0.type == 'r':
            info = getptrinfo(arg0)
            if info:
                if info.is_virtual():
                    raise InvalidLoop("promote of a virtual")
                old_guard_op = info.get_last_guard(self.optimizer)
                if old_guard_op is not None:
                    op = self.replace_old_guard_with_guard_value(op, info,
                                                              old_guard_op)
        elif arg0.type == 'f':
            arg0 = get_box_replacement(arg0)
            if arg0.is_constant():
                return
        constbox = op.getarg(1)
        assert isinstance(constbox, Const)
        return self.optimize_guard(op, constbox)

    def postprocess_GUARD_VALUE(self, op):
        box = get_box_replacement(op.getarg(0))
        self.make_constant(box, op.getarg(1))

    def replace_old_guard_with_guard_value(self, op, info, old_guard_op):
        # there already has been a guard_nonnull or guard_class or
        # guard_nonnull_class on this value, which is rather silly.
        # This function replaces the original guard with a
        # guard_value.  Must be careful: doing so is unsafe if the
        # original guard checks for something inconsistent,
        # i.e. different than what it would give if the guard_value
        # passed (this is a rare case, but possible).  If we get
        # inconsistent results in this way, then we must not do the
        # replacement, otherwise we'd put guard_value up there but all
        # intermediate ops might be executed by assuming something
        # different, from the old guard that is now removed...

        c_value = op.getarg(1)
        if not c_value.nonnull():
            raise InvalidLoop('A GUARD_VALUE(..., NULL) follows some other '
                              'guard that it is not NULL')
        previous_classbox = info.get_known_class(self.optimizer.cpu)
        if previous_classbox is not None:
            expected_classbox = self.optimizer.cpu.cls_of_box(c_value)
            assert expected_classbox is not None
            if not previous_classbox.same_constant(
                    expected_classbox):
                r = self.optimizer.metainterp_sd.logger_ops.repr_of_resop(op)
                raise InvalidLoop('A GUARD_VALUE (%s) was proven to '
                                  'always fail' % r)
        if not self.optimizer.can_replace_guards:
            return op
        descr = compile.ResumeGuardDescr()
        op = old_guard_op.copy_and_change(rop.GUARD_VALUE,
                         args=[old_guard_op.getarg(0), op.getarg(1)],
                         descr=descr)
        # Note: we give explicitly a new descr for 'op'; this is why the
        # old descr must not be ResumeAtPositionDescr (checked above).
        # Better-safe-than-sorry but it should never occur: we should
        # not put in short preambles guard_xxx and guard_value
        # on the same box.
        self.optimizer.replace_guard(op, info)
        # to be safe
        info.reset_last_guard_pos()
        return op

    def optimize_GUARD_TRUE(self, op):
        return self.optimize_guard(op, CONST_1)

    def postprocess_GUARD_TRUE(self, op):
        box = get_box_replacement(op.getarg(0))
        box1 = self.optimizer.as_operation(box, rop.INT_IS_TRUE)
        if box1 is not None:
            # we can't use the (current) range analysis for this because
            # "anything but 0" is not a valid range
            self.pure_from_args1(rop.INT_IS_ZERO, box1.getarg(0), CONST_0)
        self.make_constant(box, CONST_1)

    def optimize_GUARD_FALSE(self, op):
        return self.optimize_guard(op, CONST_0)

    def postprocess_GUARD_FALSE(self, op):
        box = get_box_replacement(op.getarg(0))
        box1 = self.optimizer.as_operation(box, rop.INT_IS_ZERO)
        if box1 is not None:
            # we can't use the (current) range analysis for this because
            # "anything but 0" is not a valid range
            self.pure_from_args1(rop.INT_IS_TRUE, box1.getarg(0), CONST_1)
        self.make_constant(box, CONST_0)

    def optimize_ASSERT_NOT_NONE(self, op):
        self.make_nonnull(op.getarg(0))

    def optimize_RECORD_EXACT_CLASS(self, op):
        opinfo = getptrinfo(op.getarg(0))
        expectedclassbox = op.getarg(1)
        assert isinstance(expectedclassbox, Const)
        if opinfo is not None:
            realclassbox = opinfo.get_known_class(self.optimizer.cpu)
            if realclassbox is not None:
                assert realclassbox.same_constant(expectedclassbox)
                return
        self.make_constant_class(op.getarg(0), expectedclassbox,
                                 update_last_guard=False)

    def optimize_record_exact_value(self, op):
        box = op.getarg(0)
        expectedconstbox = op.getarg(1)
        assert isinstance(expectedconstbox, Const)
        self.make_constant(box, expectedconstbox)

    optimize_RECORD_EXACT_VALUE_R = optimize_record_exact_value
    optimize_RECORD_EXACT_VALUE_I = optimize_record_exact_value

    def optimize_GUARD_CLASS(self, op):
        expectedclassbox = op.getarg(1)
        info = self.ensure_ptr_info_arg0(op)
        assert isinstance(expectedclassbox, Const)
        realclassbox = info.get_known_class(self.optimizer.cpu)
        if realclassbox is not None:
            if realclassbox.same_constant(expectedclassbox):
                return
            r = self.optimizer.metainterp_sd.logger_ops.repr_of_resop(op)
            raise InvalidLoop('A GUARD_CLASS (%s) was proven to always fail'
                              % r)
        old_guard_op = info.get_last_guard(self.optimizer)
        if old_guard_op and not isinstance(old_guard_op.getdescr(),
                                           compile.ResumeAtPositionDescr):
            # there already has been a guard_nonnull or guard_class or
            # guard_nonnull_class on this value.
            if (self.optimizer.can_replace_guards and
                    old_guard_op.getopnum() == rop.GUARD_NONNULL):
                # it was a guard_nonnull, which we replace with a
                # guard_nonnull_class.
                descr = compile.ResumeGuardDescr()
                op = old_guard_op.copy_and_change(rop.GUARD_NONNULL_CLASS,
                            args=[old_guard_op.getarg(0), op.getarg(1)],
                            descr=descr)
                # Note: we give explicitly a new descr for 'op'; this is why the
                # old descr must not be ResumeAtPositionDescr (checked above).
                # Better-safe-than-sorry but it should never occur: we should
                # not put in short preambles guard_nonnull and guard_class
                # on the same box.
                self.optimizer.replace_guard(op, info)
                return self.emit(op)
        return self.emit(op)

    def postprocess_GUARD_CLASS(self, op):
        expectedclassbox = op.getarg(1)
        info = getptrinfo(op.getarg(0))
        old_guard_op = info.get_last_guard(self.optimizer)
        update_last_guard = not old_guard_op or isinstance(
            old_guard_op.getdescr(), compile.ResumeAtPositionDescr)
        self.make_constant_class(op.getarg(0), expectedclassbox, update_last_guard)

    def optimize_GUARD_NONNULL_CLASS(self, op):
        info = getptrinfo(op.getarg(0))
        if info and info.is_null():
            r = self.optimizer.metainterp_sd.logger_ops.repr_of_resop(op)
            raise InvalidLoop('A GUARD_NONNULL_CLASS (%s) was proven to '
                              'always fail' % r)
        return self.optimize_GUARD_CLASS(op)

    postprocess_GUARD_NONNULL_CLASS = postprocess_GUARD_CLASS

    def optimize_CALL_LOOPINVARIANT_I(self, op):
        arg = op.getarg(0)
        # 'arg' must be a Const, because residual_call in codewriter
        # expects a compile-time constant
        assert isinstance(arg, Const)
        key = make_hashable_int(arg.getint())

        resvalue = self.loop_invariant_results.get(key, None)
        if resvalue is not None:
            resvalue = self.optimizer.force_op_from_preamble(resvalue)
            self.loop_invariant_results[key] = resvalue
            self.make_equal_to(op, resvalue)
            self.last_emitted_operation = REMOVED
            return
        # change the op to be a normal call, from the backend's point of view
        # there is no reason to have a separate operation for this
        newop = self.replace_op_with(op,
                                     OpHelpers.call_for_descr(op.getdescr()))
        return self.emit_result(CallLoopinvariantOptimizationResult(self, newop, op))

    optimize_CALL_LOOPINVARIANT_R = optimize_CALL_LOOPINVARIANT_I
    optimize_CALL_LOOPINVARIANT_F = optimize_CALL_LOOPINVARIANT_I
    optimize_CALL_LOOPINVARIANT_N = optimize_CALL_LOOPINVARIANT_I

    def optimize_COND_CALL(self, op):
        arg = op.getarg(0)
        b = self.getintbound(arg)
        if b.is_constant():
            if b.get_constant_int() == 0:
                self.last_emitted_operation = REMOVED
                return
            opnum = OpHelpers.call_for_type(op.type)
            op = op.copy_and_change(opnum, args=op.getarglist()[1:])
        return self.emit(op)

    def optimize_COND_CALL_VALUE_I(self, op):
        # look if we know the nullness of the first argument
        info = self.getnullness(op.getarg(0))
        if info == INFO_NONNULL:
            self.make_equal_to(op, op.getarg(0))
            self.last_emitted_operation = REMOVED
            return
        if info == INFO_NULL:
            opnum = OpHelpers.call_pure_for_type(op.type)
            op = self.replace_op_with(op, opnum, args=op.getarglist()[1:])
        return self.emit(op)
    optimize_COND_CALL_VALUE_R = optimize_COND_CALL_VALUE_I

    def _optimize_nullness(self, op, box, expect_nonnull):
        info = self.getnullness(box)
        if info == INFO_NONNULL:
            self.make_constant_int(op, expect_nonnull)
        elif info == INFO_NULL:
            self.make_constant_int(op, not expect_nonnull)
        else:
            return self.emit(op)

    def optimize_INT_IS_TRUE(self, op):
        if (not self.is_raw_ptr(op.getarg(0)) and
            self.getintbound(op.getarg(0)).is_bool()):
            self.make_equal_to(op, op.getarg(0))
            return
        return self._optimize_nullness(op, op.getarg(0), True)

    def optimize_INT_IS_ZERO(self, op):
        return self._optimize_nullness(op, op.getarg(0), False)

    def _optimize_oois_ooisnot(self, op, expect_isnot, instance):
        arg0 = get_box_replacement(op.getarg(0))
        arg1 = get_box_replacement(op.getarg(1))
        info0 = getptrinfo(arg0)
        info1 = getptrinfo(arg1)
        if info0 and info0.is_virtual():
            if info1 and info1.is_virtual():
                intres = (info0 is info1) ^ expect_isnot
                self.make_constant_int(op, intres)
            else:
                self.make_constant_int(op, expect_isnot)
        elif info1 and info1.is_virtual():
            self.make_constant_int(op, expect_isnot)
        elif info1 and info1.is_null():
            return self._optimize_nullness(op, op.getarg(0), expect_isnot)
        elif info0 and info0.is_null():
            return self._optimize_nullness(op, op.getarg(1), expect_isnot)
        elif arg0 is arg1:
            self.make_constant_int(op, not expect_isnot)
        else:
            if instance:
                if info0 is None:
                    cls0 = None
                else:
                    cls0 = info0.get_known_class(self.optimizer.cpu)
                if cls0 is not None:
                    if info1 is None:
                        cls1 = None
                    else:
                        cls1 = info1.get_known_class(self.optimizer.cpu)
                    if cls1 is not None and not cls0.same_constant(cls1):
                        # cannot be the same object, as we know that their
                        # class is different
                        self.make_constant_int(op, expect_isnot)
                        return
            elif isinstance(info0, ArrayPtrInfo) and isinstance(info1, ArrayPtrInfo):
                if info0.getlenbound(None).known_ne(info1.getlenbound(None)):
                    self.make_constant_int(op, expect_isnot)
                    return
            return self.emit(op)

    def optimize_PTR_EQ(self, op):
        return self._optimize_oois_ooisnot(op, False, False)

    def optimize_PTR_NE(self, op):
        return self._optimize_oois_ooisnot(op, True, False)

    def optimize_INSTANCE_PTR_EQ(self, op):
        arg0 = get_box_replacement(op.getarg(0))
        arg1 = get_box_replacement(op.getarg(1))
        self.pure_from_args2(rop.INSTANCE_PTR_EQ, arg1, arg0, op)
        return self._optimize_oois_ooisnot(op, False, True)

    def optimize_INSTANCE_PTR_NE(self, op):
        arg0 = get_box_replacement(op.getarg(0))
        arg1 = get_box_replacement(op.getarg(1))
        self.pure_from_args2(rop.INSTANCE_PTR_NE, arg1, arg0, op)
        return self._optimize_oois_ooisnot(op, True, True)

    def optimize_CALL_N(self, op):
        # dispatch based on 'oopspecindex' to a method that handles
        # specifically the given oopspec call.  For non-oopspec calls,
        # oopspecindex is just zero.
        effectinfo = op.getdescr().get_extra_info()
        oopspecindex = effectinfo.oopspecindex
        if oopspecindex == EffectInfo.OS_ARRAYCOPY:
            return self._optimize_CALL_ARRAYCOPY(op)
        if oopspecindex == EffectInfo.OS_ARRAYMOVE:
            return self._optimize_CALL_ARRAYMOVE(op)
        return self.emit(op)

    def _optimize_CALL_ARRAYCOPY(self, op):
        if self._optimize_call_arrayop(op, op.getarg(1), op.getarg(2),
                             op.getarg(3), op.getarg(4), op.getarg(5)):
            return None
        return self.emit(op)

    def _optimize_CALL_ARRAYMOVE(self, op):
        array_box = op.getarg(1)
        if self._optimize_call_arrayop(op, array_box, array_box,
                             op.getarg(2), op.getarg(3), op.getarg(4)):
            return None
        return self.emit(op)

    def _optimize_call_arrayop(self, op, source_box, dest_box,
                               source_start_box, dest_start_box, length_box):
        length = self.get_constant_box(length_box)
        if not length:
            return False
        length_int = length.getint()
        if length_int == 0:
            return True  # 0-length arraycopy or arraymove

        source_info = getptrinfo(source_box)
        dest_info = getptrinfo(dest_box)
        source_start_box = self.get_constant_box(source_start_box)
        dest_start_box = self.get_constant_box(dest_start_box)
        extrainfo = op.getdescr().get_extra_info()
        if (source_start_box and dest_start_box
            and ((dest_info and dest_info.is_virtual()) or
                            length_int <= 8) and
            ((source_info and source_info.is_virtual()) or length_int <= 8)
            and extrainfo.single_write_descr_array is not None): #<-sanity check
            source_start = source_start_box.getint()
            dest_start = dest_start_box.getint()
            arraydescr = extrainfo.single_write_descr_array
            if arraydescr.is_array_of_structs():
                # for array of structs, only support if both are virtual
                # and it's not a memmove
                if not ((source_info and source_info.is_virtual()) and
                        (dest_info and dest_info.is_virtual()) and
                        source_info is not dest_info):
                    return False
                all_fdescrs = arraydescr.get_all_fielddescrs()
                if all_fdescrs is None:
                    return False
                for index in range(length_int):
                    for fdescr in all_fdescrs:
                        box = source_info.getinteriorfield_virtual(source_start + index, fdescr)
                        dest_info.setinteriorfield_virtual(dest_start + index, fdescr, box)
                return True
            index_current = 0
            index_delta = +1
            index_stop = length_int
            if (source_box is dest_box and        # ARRAYMOVE only
                    source_start < dest_start):   # iterate in reverse order
                index_current = index_stop - 1
                index_delta = -1
                index_stop = -1

            # XXX fish fish fish
            while index_current != index_stop:
                index = index_current
                index_current += index_delta
                assert index >= 0
                if source_info and source_info.is_virtual():
                    val = source_info.getitem(arraydescr, index + source_start)
                else:
                    opnum = OpHelpers.getarrayitem_for_descr(arraydescr)
                    newop = ResOperation(opnum,
                                      [source_box,
                                       ConstInt(index + source_start)],
                                       descr=arraydescr)
                    self.optimizer.send_extra_operation(newop)
                    val = newop
                if val is None:
                    continue
                if dest_info and dest_info.is_virtual():
                    dest_info.setitem(arraydescr, index + dest_start,
                                      get_box_replacement(dest_box),
                                      val)
                else:
                    newop = ResOperation(rop.SETARRAYITEM_GC,
                                         [dest_box,
                                          ConstInt(index + dest_start),
                                          val],
                                         descr=arraydescr)
                    self.optimizer.send_extra_operation(newop)
            return True
        return False

    def optimize_CALL_PURE_I(self, op):
        # this removes a CALL_PURE with all constant arguments.
        # Note that it's also done in pure.py.  For now we need both...
        result = self._can_optimize_call_pure(op)
        if result is not None:
            self.make_constant(op, result)
            self.last_emitted_operation = REMOVED
            return

        # dispatch based on 'oopspecindex' to a method that handles
        # specifically the given oopspec call.
        effectinfo = op.getdescr().get_extra_info()
        oopspecindex = effectinfo.oopspecindex
        if oopspecindex == EffectInfo.OS_INT_UDIV:
            if self._optimize_CALL_INT_UDIV(op):
                return
        elif oopspecindex == EffectInfo.OS_INT_PY_DIV:
            if self._optimize_CALL_INT_PY_DIV(op):
                return
        elif oopspecindex == EffectInfo.OS_INT_PY_MOD:
            if self._optimize_CALL_INT_PY_MOD(op):
                return
        return self.emit(op)
    optimize_CALL_PURE_R = optimize_CALL_PURE_I
    optimize_CALL_PURE_F = optimize_CALL_PURE_I
    optimize_CALL_PURE_N = optimize_CALL_PURE_I

    def optimize_GUARD_NO_EXCEPTION(self, op):
        if self.last_emitted_operation is REMOVED:
            # it was a CALL_PURE or a CALL_LOOPINVARIANT that was killed;
            # so we also kill the following GUARD_NO_EXCEPTION
            return
        return self.emit(op)

    def optimize_GUARD_FUTURE_CONDITION(self, op):
        self.optimizer.notice_guard_future_condition(op)

    def _optimize_CALL_INT_PY_DIV(self, op):
        from rpython.jit.metainterp.optimizeopt.intutils import MININT
        def _replace_with_shift(self, op, arg1, shiftvar):
            op = self.replace_op_with(op, rop.INT_RSHIFT,
                        args=[arg1, shiftvar],
                        descr=DONT_CHANGE)  # <- xxx rename? means "kill"
            self.optimizer.send_extra_operation(op)
            return True
        arg1 = op.getarg(1)
        b1 = self.getintbound(arg1)
        arg2 = op.getarg(2)
        b2 = self.getintbound(arg2)

        if b1.known_eq_const(0):
            self.make_constant_int(op, 0)
            self.last_emitted_operation = REMOVED
            return True
        if not b2.is_constant():
            shiftop = self.optimizer.as_operation(get_box_replacement(arg2), rop.INT_LSHIFT)
            if (shiftop and
                    shiftop.getarg(0).is_constant() and
                    shiftop.getarg(0).getint() == 1):
                # x // (1 << y) == 1 >> y
                # but only if 0 <= y < LONG_BIT - 1
                shiftvar = get_box_replacement(shiftop.getarg(1))
                shiftbound = self.getintbound(shiftvar)
                if shiftbound.known_nonnegative() and shiftbound.known_lt_const(LONG_BIT - 1):
                    return _replace_with_shift(self, op, arg1, shiftvar)
            return False
        val = b2.get_constant_int()
        if val == -1:
            if b1.known_gt_const(MININT):
                op = self.replace_op_with(op, rop.INT_NEG,
                            args=[arg1],
                            descr=DONT_CHANGE)  # <- xxx rename? means "kill"
                self.optimizer.send_extra_operation(op)
                return True
        if val <= 0:
            return False
        if val == 1:
            self.make_equal_to(op, arg1)
            self.last_emitted_operation = REMOVED
            return True
        elif val & (val - 1) == 0:   # val == 2**shift
            return _replace_with_shift(self, op, arg1, ConstInt(highest_bit(val)))
        else:
            from rpython.jit.metainterp.optimizeopt import intdiv
            known_nonneg = b1.known_nonnegative()
            operations = intdiv.division_operations(arg1, val, known_nonneg)
            newop = None
            for newop in operations:
                self.optimizer.send_extra_operation(newop)
            self.make_equal_to(op, newop)
            return True

    def _optimize_CALL_INT_PY_MOD(self, op):
        arg1 = op.getarg(1)
        b1 = self.getintbound(arg1)
        arg2 = op.getarg(2)
        b2 = self.getintbound(arg2)

        if b1.known_eq_const(0):
            self.make_constant_int(op, 0)
            self.last_emitted_operation = REMOVED
            return True
        # This is Python's integer division: 'x // (2**shift)' can always
        # be replaced with 'x >> shift', even for negative values of x
        if not b2.is_constant():
            return False
        val = b2.get_constant_int()
        if val <= 0:
            return False
        if val == 1:
            self.make_constant_int(op, 0)
            self.last_emitted_operation = REMOVED
            return True
        elif val & (val - 1) == 0:   # val == 2**shift
            # x % power-of-two ==> x & (power-of-two - 1)
            # with Python's modulo, this is valid even if 'x' is negative.
            op = self.replace_op_with(op, rop.INT_AND,
                        args=[arg1, ConstInt(val - 1)],
                        descr=DONT_CHANGE)  # <- xxx rename? means "kill"
            self.optimizer.send_extra_operation(op)
            return True
        else:
            from rpython.jit.metainterp.optimizeopt import intdiv
            known_nonneg = b1.known_nonnegative()
            operations = intdiv.modulo_operations(arg1, val, known_nonneg)
            newop = None
            for newop in operations:
                self.optimizer.send_extra_operation(newop)
            self.make_equal_to(op, newop)
            return True

    def optimize_CAST_PTR_TO_INT(self, op):
        self.optimizer.pure_from_args1(rop.CAST_INT_TO_PTR, op, op.getarg(0))
        return self.emit(op)

    def optimize_CAST_INT_TO_PTR(self, op):
        self.optimizer.pure_from_args1(rop.CAST_PTR_TO_INT, op, op.getarg(0))
        return self.emit(op)

    def optimize_CONVERT_FLOAT_BYTES_TO_LONGLONG(self, op):
        self.optimizer.pure_from_args1(rop.CONVERT_LONGLONG_BYTES_TO_FLOAT, op, op.getarg(0))
        return self.emit(op)

    def optimize_CONVERT_LONGLONG_BYTES_TO_FLOAT(self, op):
        self.optimizer.pure_from_args1(rop.CONVERT_FLOAT_BYTES_TO_LONGLONG, op, op.getarg(0))
        return self.emit(op)

    def optimize_SAME_AS_I(self, op):
        self.make_equal_to(op, op.getarg(0))
    optimize_SAME_AS_R = optimize_SAME_AS_I
    optimize_SAME_AS_F = optimize_SAME_AS_I

    def serialize_optrewrite(self, available_boxes):
        res = []
        for i, box in self.loop_invariant_results.iteritems():
            box = get_box_replacement(box)
            if box in available_boxes:
                res.append((i, box))
        return res

    def deserialize_optrewrite(self, tups):
        for i, box in tups:
            self.loop_invariant_results[i] = box

dispatch_opt = make_dispatcher_method(OptRewrite, 'optimize_',
                                      default=OptRewrite.emit)
optimize_guards = _findall(OptRewrite, 'optimize_', 'GUARD')
OptRewrite.propagate_postprocess = make_dispatcher_method(OptRewrite, 'postprocess_')
OptRewrite.have_postprocess_op = have_dispatcher_method(OptRewrite, 'postprocess_')



from pypy.interpreter.mixedmodule import MixedModule

class Module(MixedModule):
    interpleveldefs = {
        'PAGESIZE': 'space.wrap(interp_mmap.PAGESIZE)',
        'ALLOCATIONGRANULARITY': 'space.wrap(interp_mmap.ALLOCATIONGRANULARITY)',
        'ACCESS_READ' : 'space.wrap(interp_mmap.ACCESS_READ)',
        'ACCESS_WRITE': 'space.wrap(interp_mmap.ACCESS_WRITE)',
        'ACCESS_COPY' : 'space.wrap(interp_mmap.ACCESS_COPY)',
        'mmap': 'interp_mmap.W_MMap',
        'error': 'space.fromcache(interp_mmap.Cache).w_error',
    }

    appleveldefs = {
    }

    def buildloaders(cls):
        from rpython.rlib import rmmap
        for constant, value in rmmap.constants.iteritems():
            if isinstance(value, (int, long)):
                Module.interpleveldefs[constant] = "space.wrap(%r)" % value
        super(Module, cls).buildloaders()
    buildloaders = classmethod(buildloaders)
